@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -68,7 +68,7 @@ Buffer *CBuf;
 Line *CLine;
 Buffer_Local_Type Buffer_Local = /*{{{*/
 {
-   8
+   8, 0, 0
 };
 
 /*}}}*/
@@ -83,6 +83,18 @@ unsigned int LineNum = 1;
 unsigned int Max_LineNum;
 int Point = 0;
 
+static void delete_buffer_hooks (Jed_Buffer_Hook_Type *);
+
+void jed_set_point (int point)
+{
+   if (point > CLine->len)
+     point = CLine->len;
+   if (point < 0)
+     point = 0;
+   
+   Point = point;
+}
+
 /* move point to top of buffer */
 int bob() /*{{{*/
 {
@@ -104,20 +116,18 @@ int eob()                      /* point to end of buffer */ /*{{{*/
 	return(0);
      }
    
-   Point = CLine->len - 1;
-   if (Point < 0)
-     {
-	Point = 0;
-	return(1);
-     }
+   if (LINE_HAS_NEWLINE(CLine)
+       && (CBuf != MiniBuffer))
+     Point = CLine->len - 1;
+   else
+     Point = CLine->len;
    
-   if ((CLine->data[Point] != '\n') || (CBuf == MiniBuffer)) Point++;
-   return(1);
+   return 1;
 }
 
 /*}}}*/
 
-int bol() /*{{{*/
+int bol(void) /*{{{*/
 {
    Point = 0;
    Goal_Column = 1;
@@ -128,195 +138,439 @@ int bol() /*{{{*/
 
 int eol (void) /*{{{*/
 {
-   int nl = 1;
-   if (CLine != NULL)
+   if (CLine == NULL)
      {
-	Point = CLine->len - 1;
-	if (Point < 0)
-	  {
-	     Point = 0;
-	     return 0;
-	  }
-	
-	if (((CLine->data)[Point] != '\n') || (CBuf == MiniBuffer)) nl = 0;
-	if (!nl) Point++;
+	Point = 0;
+	return 1;
      }
-   
-   else Point = 0;
+
+   if (LINE_HAS_NEWLINE(CLine)
+       && (CBuf != MiniBuffer))
+     Point = CLine->len - 1;
+   else
+     Point = CLine->len;
+
    return 1;
 }
 
 /*}}}*/
 
-int bobp()                /* returns 1 if top line of buffer, 0 otherwise */ /*{{{*/
+int bobp (void)                /* returns 1 if top line of buffer, 0 otherwise */ /*{{{*/
 {
-   if ((CBuf->beg == CLine) && (Point == 0)) return(1); else return(0);
+   return ((CBuf->beg == CLine) && (Point == 0));
 }
 
 /*}}}*/
 
-int eolp() /*{{{*/
+int eolp (void) /*{{{*/
 {
    int len;
    
-   if (CLine == NULL) return(1);
-   len = CLine->len;
-   if (len == 0) return(1);
-   if (Point < len - 1) return(0);
+   if ((CLine == NULL)
+       || (0 == (len = CLine->len)))
+     return 1;
+
+   if (Point < len - 1) return 0;
    
    if (CBuf == MiniBuffer)
-     {
-	if (Point < len) return 0; else return (1);
-     }
-   
-   
-   /* Point is either len or len - 1 */
-   
-   if ((CLine->data)[len-1] == '\n') return (1);
-   
-   if (Point == len) return(1);
+     return (Point >= len);
+
+   /* Point is either len or len - 1.  If it is len-1, then we are at the 
+    * end of the line only if the line ends with \n.
+    */
+
+   if (LINE_HAS_NEWLINE(CLine)
+       || (Point == len))
+     return 1;
+
    Point = len - 1;
-   return(0);
+   return 0;
 }
 
 /*}}}*/
 
-int eobp() /*{{{*/
+int eobp (void) /*{{{*/
 {
-   if (CLine != CBuf->end) return(0);
-   return(eolp());
+   return ((CLine == CBuf->end) && eolp ());
 }
 
 /*}}}*/
 
-int bolp() /*{{{*/
+int bolp (void) /*{{{*/
 {
-   if (Point) return(0); else return(1);
+   return (Point == 0);
 }
 
 /*}}}*/
 
 /*  Attempt to goback n lines, return actual number. */
-int prevline(int *n) /*{{{*/
+unsigned int jed_up (unsigned int n) /*{{{*/
 {
-   int i = 0;
-   Line *prev;
-   
-   Point = 0;     /* return to beginning of this line */
-   while (i < *n)
+   unsigned int i = 0;
+
+   if (n == 0)
+     return 0;
+
+   while (i < n)
      {
-	prev = CLine->prev;
-	while((prev != NULL) && (prev->len == 0)) prev = prev->prev;
+	Line *prev = CLine->prev;
+	
+	/* Why would prev->len be 0 ??? */
+	while ((prev != NULL) && (prev->len == 0)) 
+	  prev = prev->prev;
+
 	if (prev == NULL) break;
 	i++;
 	CLine = prev;
      }
-   
-   if (i) eol();
-   LineNum -= (unsigned int) i;
-   return(i);
+   LineNum -= i;
+
+   if (i) eol ();
+   return i;
 }
 
 /*}}}*/
 
-int nextline(int *np) /*{{{*/
+unsigned int jed_down (unsigned int n) /*{{{*/
 {
-   register Line *next;
-   register int i = 0, n = *np;
-   
-   while(i < n)
+   unsigned int i = 0;
+
+   while (i < n)
      {
-	next = CLine->next;
+	Line *next = CLine->next;
 	if (next == NULL) break;
 	CLine = next;
 	i++;
      }
-   
-   if (i) Point = 0;
-   LineNum += (unsigned int) i;
-   return(i);
+   LineNum += i;
+
+   if (i) bol ();
+   return i;
 }
 
 /*}}}*/
 
+unsigned char *jed_multibyte_chars_forward (unsigned char *p, unsigned char *pmax,
+					    unsigned int n, unsigned int *dn, 
+                                            int ignore_combining)
+{
+   (void) ignore_combining;
+#if JED_HAS_UTF8_SUPPORT
+   if (Jed_UTF8_Mode == 0)
+     {
+#endif
+	if (p + n > pmax)
+	  n = pmax - p;
+   
+	p += n;
+	if (dn != NULL) *dn = n;
+	return p;
+#if JED_HAS_UTF8_SUPPORT
+     }
+   return SLutf8_skip_chars (p, pmax, n, dn, ignore_combining);
+#endif
+}
+
+unsigned char *jed_multibyte_chars_backward (unsigned char *pmin, unsigned char *p, 
+					     unsigned int n, unsigned int *dn, 
+                                             int ignore_combining)
+{
+   (void) ignore_combining;
+#if JED_HAS_UTF8_SUPPORT
+   if (Jed_UTF8_Mode == 0)
+     {
+#endif
+	if (pmin + n > p)
+	  n = p - pmin;
+   
+	p -= n;
+	if (dn != NULL) *dn = n;
+	return p;
+#if JED_HAS_UTF8_SUPPORT
+     }
+   return SLutf8_bskip_chars (pmin, p, n, dn, ignore_combining);
+#endif
+}
+
+int jed_multibyte_charcasecmp (unsigned char **ap, unsigned char *amax,
+			       unsigned char **bp, unsigned char *bmax)
+{
+   SLuchar_Type *a, *b;
+   SLwchar_Type cha, chb;
+
+   a = *ap;
+   b = *bp;
+   
+   if ((a >= amax) || (b >= bmax))
+     return 0;
+
+#if JED_HAS_UTF8_SUPPORT
+   if (Jed_UTF8_Mode)
+     {
+        unsigned int na, nb;
+        int aok, bok;
+
+        aok = (NULL != SLutf8_decode (a, amax, &cha, &na));
+        bok = (NULL != SLutf8_decode (b, bmax, &chb, &nb));
+        
+        *ap = a + na;
+        *bp = b + nb;
+        
+        if (aok && bok)
+          {
+             cha = SLwchar_toupper (cha);
+             chb = SLwchar_toupper (chb);
+          }
+        else if (aok)
+          return 1;
+        else if (bok)
+          return -1;
+        
+# if SLANG_VERSION < 20005
+	/* Earlier versions of SLutf8_decode did not set the value of the
+	 * SLwchar_Type arg to the byte-value
+	 */
+	cha = *a;
+	chb = *b;
+# endif
+        if (cha > chb)
+          return 1;
+        
+        if (cha == chb)
+          return 0;
+        
+        return -1;
+     }
+#endif
+
+   *ap = a+1;
+   *bp = b+1;
+
+   if ((*a == *b) || ((cha = UPPER_CASE(*a)) == (chb = UPPER_CASE(*b))))
+     return 0;
+   if (cha > chb) return 1;
+   return -1;
+}
+
+/* The buffer is required to be at least JED_MAX_MULTIBYTE_SIZE bytes wide */
+unsigned char *jed_wchar_to_multibyte (SLwchar_Type c, unsigned char *buf)
+{
+#if JED_HAS_UTF8_SUPPORT
+   if (Jed_UTF8_Mode && (c >= 0x80))
+     {
+	unsigned char *b = SLutf8_encode (c, buf, JED_MAX_MULTIBYTE_SIZE);
+	if (b == NULL)
+	  SLang_verror (SL_Unicode_Error, "Unable to convert 0x%lX to UTF-8", (unsigned long)c);
+	return b;
+     }
+#endif
+   *buf++ = c;
+   return buf;
+}
+
+int jed_what_char (SLwchar_Type *w) /*{{{*/
+{
+#if JED_HAS_UTF8_SUPPORT
+   SLuchar_Type *p, *pmax;
+
+   if (eobp ())
+     {
+	*w = 0;
+	return -1;
+     }
+   p = CLine->data + Point;
+   if (Jed_UTF8_Mode == 0)
+     {
+        *w = (SLwchar_Type) *p;
+        return 0;
+     }
+   pmax = SLutf8_skip_char (p, CLine->data + CLine->len);
+   if (NULL == SLutf8_decode (p, pmax, w, NULL))
+     {
+	if (p < pmax)
+	  *w = (SLwchar_Type) *p;
+	else
+	  *w = 0;
+	return -1;
+     }
+
+   return 0;
+#else
+   if (eobp ())
+     return -1;
+   *w = (SLwchar_Type) *(CLine->data + Point);
+   return 0;
+#endif
+}
+
+/*}}}*/
+
+
+void jed_count_chars (void) /*{{{*/
+{
+   unsigned long n = 0, m = 0;
+   char buf[256];
+   SLwchar_Type w;
+
+   Line *l = CBuf->beg;
+   
+   while (l != NULL)
+     {
+	n += l->len;
+	l = l->next;
+     }
+   l = CBuf->beg;
+   while (l != CLine)
+     {
+	m += l->len;
+	l = l->next;
+     }
+   m += Point + 1;
+
+   if (eobp ())
+     *buf = 0;
+   else
+     {
+        if (-1 == jed_what_char (&w))
+	  sprintf (buf, "-<%0X>, ", (int) w);
+        else
+          {
+             SLuchar_Type *p, *pmax;
+             
+             p = CLine->data + Point;
+             if (*p == 0)
+               {
+                  p = (SLuchar_Type *) "^@";
+                  pmax = p + 2;
+               }
+             else
+               pmax = jed_multibyte_chars_forward (p, CLine->data + CLine->len, 1, NULL, 1);
+             
+             buf[0] = '\'';
+             strncpy (buf + 1, (char *)p, pmax - p);
+             p = (SLuchar_Type *) (buf + 1 + (pmax - p));
+
+             sprintf ((char *)p, "'=%ld/0x%lX/%#lo, ", (long)w, (unsigned long)w, (unsigned long)w);
+          }
+     }
+   
+   sprintf (buf + strlen (buf), "point %lu of %lu", m, n);
+   SLang_push_string(buf);
+}
+
+/*}}}*/
+
+
+unsigned char *jed_eol_position (Line *l)
+{
+   unsigned char *p = l->data + l->len;
+   if (LINE_HAS_NEWLINE(l))
+     p--;
+   
+   return p;
+}
+
+void jed_position_point (unsigned char *p)
+{
+   if (p < CLine->data + CLine->len)
+     {
+	Point = p - CLine->data;
+	if (Point < 0)
+	  Point = 0;
+	return;
+     }
+   eol ();
+}
+
 /* The algorithm:  go to Point = len - 1.  Then forward a line counted as 1 */
-int forwchars(int *np) /*{{{*/
+unsigned int jed_right (unsigned int n) /*{{{*/
 {
-   int len, total = 0;
-   unsigned char *p;
-   Line *next;
-   int n = *np;
+   unsigned int total = 0;
    
-   if (n < 0) return 0;
-   if (CBuf == MiniBuffer)
+   while (1)
      {
-	total = CLine->len - Point;
-	if (n > total) n = total;
-	Point += n;
-	return n;
-     }
-   
-   while(1)
-     {
-	len = CLine->len;
-	/* if (Point == len) return(total); */
-	p = CLine->data + Point;
-	while(n && (Point < len) && (*p != '\n'))
+	int len = CLine->len;
+	unsigned char *p = CLine->data + Point;
+	unsigned char *pmax = CLine->data + len;
+	int has_newline = LINE_HAS_NEWLINE(CLine);
+	unsigned int dn;
+
+	if (has_newline)
+	  pmax--;
+
+	p = jed_multibyte_chars_forward (p, pmax, n, &dn, 1);
+	Point = p - CLine->data;
+	n -= dn;
+	total += dn;
+
+	if (n == 0)
+	  return total;
+
+	/* skip over the newline */
+	if (has_newline)
 	  {
 	     Point++;
-	     n--;
-	     p++;
-	     total++;
-	  }
-	if (!n) return(total);
-	if ((Point < len) && (*p == '\n'))
-	  {
 	     total++;
 	     n--;
 	  }
-	
-	if (NULL == (next = CLine->next))
+
+	if (0 == jed_down (1))
 	  {
-	     if ((Point < len) && (*p == '\n')) total--;
-	     return(total);
+	     if (has_newline && (CBuf != MiniBuffer))
+	       {
+		  Point--;
+		  total--;
+	       }
+	     return total;
 	  }
-	
-	Point = 0;
-	CLine = next;
-	LineNum++;
      }
 }
 
 /*}}}*/
 
-int backwchars(int *np) /*{{{*/
+unsigned int jed_right_bytes (unsigned int n)
 {
-   int total = 0;
-   int n = *np;
+#if JED_HAS_UTF8_SUPPORT
+   int mode = Jed_UTF8_Mode;
+   Jed_UTF8_Mode = 0;
+#endif
+   n = jed_right (n);
+#if JED_HAS_UTF8_SUPPORT
+   Jed_UTF8_Mode = mode;
+#endif
+   return n;
+}
+
+/*}}}*/
+
+unsigned int jed_left (unsigned int n) /*{{{*/
+{
+   unsigned int total = 0;
    
-   if (n < 0) return 0;
-   if (CBuf == MiniBuffer)
+   while (1)
      {
-	if (n > Point) n = Point;
-	Point -= n;
-	return (n);
-     }
-   
-   
-   do
-     {
-	if (n <= Point)
+	unsigned char *pmin = CLine->data;
+	unsigned char *p = pmin + Point;
+	unsigned int dn;
+
+	p = jed_multibyte_chars_backward (pmin, p, n, &dn, 1);
+
+	total += dn;
+	n -= dn;
+	Point = p - pmin;
+
+	if (n == 0)
+	  return total;
+
+	if (0 == jed_up (1))
+	  return total;
+
+	if (LINE_HAS_NEWLINE(CLine))
 	  {
-	     total += n;
-	     Point = Point - n;
-	     return(total);
+	     n--;
+	     total++;
 	  }
-	n = n - Point - 1;   /* Point + newline */
-	total += Point + 1;   /* ditto */
      }
-   while(prevline(&Number_One));
-   
-   return(total - 1);
 }
 
 /*}}}*/
@@ -764,7 +1018,8 @@ Buffer *make_buffer (char *name, char *dir, char *file) /*{{{*/
    newB->keymap = Global_Map;
    jed_set_buffer_ctime (newB);
    newB->local_vars.tab = Jed_Tab_Default;
-   
+   newB->local_vars.case_search = Jed_Case_Search_Default;
+
    if (CBuf == NULL)
      {
 	newB->next = newB;
@@ -796,6 +1051,7 @@ Buffer *make_buffer (char *name, char *dir, char *file) /*{{{*/
 
 /*}}}*/
 
+
 /* if there is a window attached to this buffer, then there are problems
  *  if we get to update() without attaching another one to it.  So
  *  beware! Always make sure CBuf is set too!   kill_buffer command
@@ -805,6 +1061,8 @@ void delete_buffer(Buffer *buf) /*{{{*/
    Line *l,*n;
    Jed_Mark_Array_Type *m, *m1;
 
+   if (1 != buffer_exists (buf))
+     return;
 
    jed_widen_whole_buffer (buf);
 
@@ -862,10 +1120,12 @@ void delete_buffer(Buffer *buf) /*{{{*/
    
    SLang_free_slstring (buf->dir);
    SLang_free_slstring (buf->file);
+   SLfree (buf->canonical_dirfile);
    SLfree (buf->dirfile);
    SLang_free_slstring (buf->name);
    SLang_free_slstring (buf->mode_string);
-
+   
+   delete_buffer_hooks (buf->buffer_hooks);
    SLfree ((char *)buf);
 }
 
@@ -876,36 +1136,27 @@ void goto_line (int *np) /*{{{*/
    unsigned int n;
    unsigned int half1 = LineNum / 2;
    unsigned int half2 = (Max_LineNum + LineNum) / 2;
-   int sn;
    
    if (*np <= 1) n = 0; else n = (unsigned int) *np;
-   
+
    if (n < LineNum)
      {
 	if (n > half1)
-	  {
-	     sn = (int) (LineNum - n);
-	     prevline(&sn);
-	  }
+	  jed_up (LineNum - n);
 	else
 	  {
 	     bob();
-	     sn = (int) n;
-	     if (sn-- > 0) nextline(&sn);
+	     if (n > 0) jed_down (n-1);
 	  }
      }
    else if (n < half2)
-     {
-	sn = (int) (n - LineNum);
-	nextline(&sn);
-     }
+     jed_down (n - LineNum);
    else
      {
 	eob();
-	sn = (int) (Max_LineNum - n);
-	prevline(&sn);
+	if (n < Max_LineNum) jed_up (Max_LineNum - n);
      }
-   Point = 0;
+   bol ();
 }
 
 /*}}}*/
@@ -972,6 +1223,10 @@ Buffer *find_buffer(char *name) /*{{{*/
 int buffer_exists(Buffer *b) /*{{{*/
 {
    Buffer *c = CBuf;
+
+   if (b == NULL)
+     return 0;
+
    do
      {
 	if (b == c) return 1;
@@ -1030,23 +1285,48 @@ int switch_to_buffer(Buffer *buf) /*{{{*/
 
 /*}}}*/
 
-#if 1
+#ifdef REAL_UNIX_SYSTEM
+static Buffer *find_file_buffer_via_inode (char *file)
+{
+   int device, inode;
+   Buffer *b;
+
+   if (-1 == jed_get_inode_info (file, &device, &inode))
+     return NULL;
+   
+   if ((device == -1) || (inode == -1))
+     return NULL;
+
+   b = CBuf;
+   do
+     {
+	if ((inode == b->inode) && (device == b->device))
+	  return b;
+
+	b = b->next;
+     }
+   while (b != CBuf);
+   
+   return NULL;
+}
+#endif
+
 Buffer *find_file_buffer(char *file) /*{{{*/
 {
    Buffer *b;
 
-   file = jed_expand_link (file);
+   file = jed_get_canonical_pathname (file);
    if (file == NULL)
      return NULL;
-   
+
    b = CBuf;
    do
      {
-	if ((b->dirfile != NULL)
+	if ((b->canonical_dirfile != NULL)
 # if JED_FILE_PRESERVE_CASE
-	    && (0 == strcmp(b->dirfile, file))
+	    && (0 == strcmp(b->canonical_dirfile, file))
 # else
-	    && (0 == jed_case_strcmp (b->dirfile, file))
+	    && (0 == jed_case_strcmp (b->canonical_dirfile, file))
 # endif
 	    )
 	  {
@@ -1056,78 +1336,31 @@ Buffer *find_file_buffer(char *file) /*{{{*/
 	b = b->next;
      }
    while (b != CBuf);
+
+# ifdef REAL_UNIX_SYSTEM
+   if (NULL != (b = find_file_buffer_via_inode (file)))
+     {
+	if (0 != strcmp (file, b->canonical_dirfile))
+	  jed_vmessage (0, "%s and %s are the same", file, b->canonical_dirfile);
+	SLfree (file);
+	return b;
+     }
+# endif
+   
    SLfree (file);
    return NULL;
 }
-
-#else
-/* we assume that file has already been expanded--- that is, canonical. */
-Buffer *find_file_buffer(char *file) /*{{{*/
-{
-   Buffer *b;
-   char *f;
-   long n, m;
-   char *dirfile;
-#ifdef REAL_UNIX_SYSTEM
-   int inode, device;
-   
-   get_inode_info (file, &device, &inode);
-#endif
-
-   f = extract_file(file);
-   n = f - file;
-   
-   b = CBuf;
-   do
-     {
-#ifdef REAL_UNIX_SYSTEM
-	if ((device != -1) && (inode != -1)
-	    && (b->device != -1) && (b->inode != -1))
-	  {
-	     if ((inode == b->inode) 
-		 && (device == b->device)
-		 && !strcmp(b->file, f)
-		 )
-	       return b;
-	  }
-	else 
-	  {
-	     m = strlen(b->dir);
-	     if ((m == n) && (!strcmp(b->file,f))
-		 && (!strncmp(b->dir, file, (unsigned int) n)))
-	       return b;
-	  }
-#else
-	m = strlen(b->dir);
-	
-	if ((m == n) 
-# if JED_FILE_PRESERVE_CASE
-	    && (0 == strcmp(b->file, f))
-	    && (0 == strncmp(b->dir, file, (unsigned int) n))
-# else
-	    && (0 == jed_case_strcmp (b->file,f))
-	    && (0 == jed_case_strncmp (b->dir, file, (unsigned int) n))
-# endif
-	    )
-	  return b;
-#endif
-	b = b->next;
-     }
-   while (b != CBuf);
-   
-   return NULL;
-}
 /*}}}*/
-#endif
 
 /* take a dir and a filename, expand them then put in buffer structure */
 void buffer_filename(Buffer *b, char *dir, char *file) /*{{{*/
 {
    unsigned int dirlen;
    char *dirfile;
+   char *canonical_dirfile;
 
    if (dir == NULL)
-     dir = get_cwd ();
+     dir = jed_get_cwd ();
 
    if (file == NULL)
      {
@@ -1139,17 +1372,17 @@ void buffer_filename(Buffer *b, char *dir, char *file) /*{{{*/
    
    dirfile = jed_dir_file_merge (dir, file);
    if (dirfile != NULL)
-     {
-	char *tmp = dirfile;
-	dirfile = jed_expand_link (dirfile);
-	SLfree (tmp);
-     }
+     canonical_dirfile = jed_get_canonical_pathname (dirfile);
+   else 
+     canonical_dirfile = NULL;
 
    SLfree (b->dirfile);
+   SLfree (b->canonical_dirfile);
    SLang_free_slstring (b->file);
    SLang_free_slstring (b->dir);
    
    if ((NULL == (b->dirfile = dirfile))
+       || (NULL == (b->canonical_dirfile = canonical_dirfile))
        || (NULL == (b->file = SLang_create_slstring (file)))
        || (NULL == (b->dir = SLang_create_nslstring (dir, dirlen))))
      {
@@ -1157,7 +1390,7 @@ void buffer_filename(Buffer *b, char *dir, char *file) /*{{{*/
      }
 
 #ifdef REAL_UNIX_SYSTEM
-   get_inode_info (b->dir, &b->device, &b->inode);
+   jed_get_inode_info (b->canonical_dirfile, &b->device, &b->inode);
 #endif
 
    if (*file)
@@ -1172,7 +1405,7 @@ int erase_buffer (void) /*{{{*/
 {
    Line *beg, *tthis;
 
-   if (-1 == jed_prepare_for_modification ())
+   if (-1 == jed_prepare_for_modification (0))
      return -1;
 
    /* CLine = CBuf->end; */
@@ -1230,7 +1463,13 @@ void mark_buffer_modified (Buffer *b, int modif, int with_lock) /*{{{*/
 void jed_set_buffer_flags (Buffer *b, unsigned int flags)
 {
    unsigned int modif;
-   
+
+#if 1
+   if (((flags & FILE_MODIFIED) == 0)
+       && (b->flags & FILE_MODIFIED))
+     jed_set_buffer_ctime (b);
+#endif
+
    modif = flags & BUFFER_MODIFIED;
    b->flags &= BUFFER_MODIFIED;
    b->flags |= flags & ~BUFFER_MODIFIED;
@@ -1258,11 +1497,19 @@ void check_line() /*{{{*/
 /*}}}*/
 
 
-static SLang_Name_Type **find_buffer_hook (Buffer *b, char *s)
+static SLang_Name_Type **find_buffer_hook (Jed_Buffer_Hook_Type *b, char *s)
 {
+   if (b == NULL)
+     return NULL;
+
    if (0 == strcmp("par_sep", s)) return &b->par_sep;
+   if (0 == strcmp("format_paragraph_hook", s)) return &b->format_paragraph_hook;
+   if (0 == strcmp("mark_paragraph_hook", s)) return &b->mark_paragraph_hook;
+   if (0 == strcmp("forward_paragraph_hook", s)) return &b->forward_paragraph_hook;
+   if (0 == strcmp("backward_paragraph_hook", s)) return &b->backward_paragraph_hook;
    if (0 == strcmp("indent_hook", s)) return &b->indent_hook;
    if (0 == strcmp("wrap_hook", s)) return &b->wrap_hook;
+   if (0 == strcmp("wrapok_hook", s)) return &b->wrapok_hook;
    if (0 == strcmp("newline_indent_hook", s)) return &b->newline_indent_hook;
    if (0 == strcmp("bob_eob_error_hook", s)) return &b->bob_eob_error_hook;
 #ifdef HAS_MOUSE
@@ -1275,16 +1522,20 @@ static SLang_Name_Type **find_buffer_hook (Buffer *b, char *s)
 # endif
 #endif
    if (0 == strcmp("update_hook", s)) return &b->update_hook;
+   if (0 == strcmp("color_region_hook", s)) return &b->color_region_hook;
 
    return NULL;
 }
 
 
-int jed_unset_buffer_hook (Buffer *b, char *name)
+int jed_unset_buffer_hook (Buffer *buffer, char *name)
 {
-   if (b == NULL) 
+   Jed_Buffer_Hook_Type *b;
+
+   if ((buffer == NULL)
+       || (NULL == (b = buffer->buffer_hooks)))
      return -1;
-   
+
    if ((name != NULL) && (*name != 0))
      {
 	SLang_Name_Type **nt = find_buffer_hook (b, name);
@@ -1294,41 +1545,65 @@ int jed_unset_buffer_hook (Buffer *b, char *name)
 	*nt = NULL;
 	return 0;
      }
-
-   /* Otherwise, unset all of them */
-
-   b->par_sep = NULL;
-   b->indent_hook = NULL;
-   b->wrap_hook = NULL;
-   b->newline_indent_hook = NULL;
-   b->bob_eob_error_hook = NULL;
-   b->update_hook = NULL;
-
-#ifdef HAS_MOUSE
-   b->mouse_down_hook = NULL;
-   b->mouse_up_hook = NULL;
-   b->mouse_drag_hook = NULL;
-# if JED_HAS_MULTICLICK
-   b->mouse_2click_hook = NULL;
-   b->mouse_3click_hook = NULL;
-# endif
-#endif
    
+   memset ((char *)b, 0, sizeof (Jed_Buffer_Hook_Type));
    return 0;
 }
 
-int jed_set_buffer_hook (Buffer *b, char *name, SLang_Name_Type *nt)
+int jed_set_buffer_hook (Buffer *buffer, char *name, SLang_Name_Type *nt)
 {
    SLang_Name_Type **ntp;
+   Jed_Buffer_Hook_Type *b;
+   
+   if (buffer == NULL)
+     return -1;
+   
+   b = buffer->buffer_hooks;
 
    if (b == NULL)
-     return -1;
-
+     {
+	b = (Jed_Buffer_Hook_Type *)SLmalloc (sizeof (Jed_Buffer_Hook_Type));
+	if (b == NULL)
+	  return -1;
+	memset ((char *) b, 0, sizeof (Jed_Buffer_Hook_Type));
+	buffer->buffer_hooks = b;
+     }
+   
    ntp = find_buffer_hook (b, name);
    if (ntp == NULL)
      return -1;
    
+   SLang_free_function (*ntp);
    *ntp = nt;
    return 0;
 }
 
+static void delete_buffer_hooks (Jed_Buffer_Hook_Type *h)
+{
+   if (h == NULL)
+     return;
+
+   if (h->mark_paragraph_hook != NULL) SLang_free_function (h->mark_paragraph_hook);
+   if (h->format_paragraph_hook != NULL) SLang_free_function (h->format_paragraph_hook);
+   if (h->forward_paragraph_hook != NULL) SLang_free_function (h->forward_paragraph_hook);
+   if (h->backward_paragraph_hook != NULL) SLang_free_function (h->backward_paragraph_hook);
+   if (h->par_sep != NULL) SLang_free_function (h->par_sep);
+   if (h->indent_hook != NULL) SLang_free_function (h->indent_hook);
+   if (h->newline_indent_hook != NULL) SLang_free_function (h->newline_indent_hook);
+   if (h->wrap_hook != NULL) SLang_free_function (h->wrap_hook);
+   if (h->wrapok_hook != NULL) SLang_free_function (h->wrapok_hook);
+   if (h->bob_eob_error_hook != NULL) SLang_free_function (h->bob_eob_error_hook);
+#ifdef HAS_MOUSE
+   if (h->mouse_down_hook != NULL) SLang_free_function (h->mouse_down_hook);
+   if (h->mouse_up_hook != NULL) SLang_free_function (h->mouse_up_hook);
+   if (h->mouse_drag_hook != NULL) SLang_free_function (h->mouse_drag_hook);
+# if JED_HAS_MULTICLICK
+   if (h->mouse_2click_hook != NULL) SLang_free_function (h->mouse_2click_hook);
+   if (h->mouse_3click_hook != NULL) SLang_free_function (h->mouse_3click_hook);
+# endif
+#endif
+   if (h->update_hook != NULL) SLang_free_function (h->update_hook);
+   if (h->color_region_hook != NULL) SLang_free_function (h->color_region_hook);
+
+   SLfree ((char *) h);
+}
