@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -13,9 +13,6 @@
 #ifdef __WIN32__
 /* This needs to go first before stdio is included. */
 # include <windows.h>
-# if !defined(__MINGW32__) && !defined(__CYGWIN32__)
-#  define sleep Sleep
-# endif
 # include <fcntl.h>
 # include <io.h>
 # include <sys/stat.h>
@@ -119,6 +116,8 @@ extern unsigned int sleep (unsigned int);
 #include "ledit.h"
 #include "userinfo.h"
 #include "hooks.h"
+#include "screen.h"
+#include "sig.h"
 
 /*}}}*/
 
@@ -463,7 +462,9 @@ int insert_file_pointer(VFILE *vp) /*{{{*/
    while(NULL != (vbuf = (unsigned char *) vgets(vp, &num)))
      {
 	n++;
-	if (SLang_Error) break;
+	if (SLang_get_error ())
+	  break;
+
 	if (-1 == jed_quick_insert (vbuf, (int) num))
 	  return -1;
      }
@@ -540,13 +541,19 @@ static int jed_write1(int fd, char *b, unsigned int n) /*{{{*/
 	  {
 #ifdef EINTR
 	     if (errno == EINTR)
-	       continue;
+	       {
+		  if (0 == jed_handle_interrupt ())
+		    continue;
+	       }
 #endif
 #ifdef EAGAIN
 	     if (errno == EAGAIN)
 	       {
-		  sleep (1);
-		  continue;
+		  if (0 == jed_handle_interrupt ())
+		    {
+		       jed_sleep (1);
+		       continue;
+		    }
 	       }
 #endif
 #ifdef ENOSPC
@@ -719,7 +726,7 @@ int write_region_to_fp (int fp) /*{{{*/
    if (!check_region(&Number_One)) return(-1);
    last = CLine; last_pnt = Point;
 
-   pop_mark(&Number_One);
+   jed_pop_mark(1);
    first = CLine; pnt = Point;
 
    /* first should never be null without hitting last first.  If this
@@ -735,12 +742,12 @@ int write_region_to_fp (int fp) /*{{{*/
 	/* This goes here inside the loop because it is possible for external
 	   events to set error_buffer */
 	pnt = 0;
-	if (SLang_Error) break;
+	if (SLang_get_error ()) break;
 	first = first->next;
 	n++;
      }
 
-   if (!SLang_Error && (last_pnt != 0))
+   if (!SLang_get_error () && (last_pnt != 0))
      {
 	len = last_pnt - pnt;
 	if (len != jed_write(fp, (char *) (last->data + pnt), len))
@@ -760,7 +767,7 @@ int write_region_to_fp (int fp) /*{{{*/
    /* Now flush output buffer if necessary */
    
    len = (int) (Output_Bufferp - Output_Buffer);
-   if (!SLang_Error && len) if (len != jed_write1(fp, Output_Buffer, len))
+   if (!SLang_get_error () && len) if (len != jed_write1(fp, Output_Buffer, len))
      {
 	msg_error(err);
      }
@@ -770,7 +777,7 @@ int write_region_to_fp (int fp) /*{{{*/
    
    pop_spot();
    VFile_Mode = VFILE_TEXT;
-   if (SLang_Error) return(-1);
+   if (SLang_get_error ()) return(-1);
    return(n);
 }
 
@@ -784,9 +791,12 @@ static int jed_close (int fp) /*{{{*/
 #ifndef IBMPC_SYSTEM
 	if (errno == EINTR) 
 	  {
-	     errno = 0;
-	     sleep (1);
-	     continue;
+	     if (-1 == jed_handle_interrupt ())
+	       {
+		  errno = 0;
+		  jed_sleep (1);
+		  continue;
+	       }
 	  }
 #endif
 #endif
@@ -891,7 +901,7 @@ static int write_file_internal (char *file, int omode) /*{{{*/
 #endif
    
    bob();
-   push_mark();  m = CBuf->marks;
+   jed_push_mark();  m = CBuf->marks;
    eob();
    fnl = Require_Final_Newline;
    if (CBuf->flags & BINARY_FILE) 
@@ -909,7 +919,7 @@ static int write_file_internal (char *file, int omode) /*{{{*/
 
    Require_Final_Newline = fnl;
    VFile_Mode = VFILE_TEXT;
-   if (m == CBuf->marks) pop_mark(&Number_Zero);
+   if (m == CBuf->marks) jed_pop_mark(0);
    
 #if JED_HAS_SAVE_NARROW
    jed_pop_narrow ();
@@ -951,12 +961,12 @@ static int make_autosave_filename(char *save, unsigned int buflen, char *dir, ch
    else
      {
 #if defined(IBMPC_SYSTEM)
-	_SLsnprintf (save, buflen, "%s#%s", dir, file);
+	SLsnprintf (save, buflen, "%s#%s", dir, file);
 #else
 # ifdef VMS
-	_SLsnprintf (save, buflen, "%s_$%s;1", dir, file);
+	SLsnprintf (save, buflen, "%s_$%s;1", dir, file);
 # else
-	_SLsnprintf (save, buflen, "%s#%s#", dir, file);
+	SLsnprintf (save, buflen, "%s#%s#", dir, file);
 # endif
 #endif
      }
@@ -968,8 +978,9 @@ static int make_autosave_filename(char *save, unsigned int buflen, char *dir, ch
 #ifndef VMS
 int jed_copy_file (char *from, char *to) /*{{{*/
 {
-   int mode;
-   short uid, gid;
+   mode_t mode;
+   uid_t uid;
+   gid_t gid;
    FILE *fp0, *fp1;
    char buf[0x7FFF];
    unsigned int readlen;
@@ -1103,8 +1114,10 @@ int write_file_with_backup(char *dirfile) /*{{{*/
 {
    char autosave_file[JED_MAX_PATH_LEN];
    int n;
-   int mode, do_mode;
-   short uid, gid;
+   int do_mode;
+   uid_t uid;
+   gid_t gid;
+   mode_t mode;
 #ifndef VMS
    char *old = NULL;
    int backup_went_ok;
@@ -1227,14 +1240,14 @@ void auto_save_buffer(Buffer *b) /*{{{*/
        || (0 == (b->flags & (AUTO_SAVE_BUFFER | AUTO_SAVE_JUST_SAVE))))
      return;
 
-   if (b->dirfile == NULL)
+   if (b->canonical_dirfile == NULL)
      return;
 
 #if !JED_HAS_SAVE_NARROW
    if (b->narrow != NULL) return;
 #endif
 
-   if (-1 == jed_dirfile_to_dir_file (b->dirfile, &dir, &file))
+   if (-1 == jed_dirfile_to_dir_file (b->canonical_dirfile, &dir, &file))
      return;
 
    old_buf = CBuf;
@@ -1253,7 +1266,7 @@ void auto_save_buffer(Buffer *b) /*{{{*/
 	     message("autosaving...done");
 	  }
       }
-   else (void) write_file_with_backup(b->dirfile);
+   else (void) write_file_with_backup(b->canonical_dirfile);
 
    SLfree (file);
    SLfree (dir);
@@ -1346,10 +1359,10 @@ static int is_link(char *f, char *f1) /*{{{*/
 
 
 #ifdef REAL_UNIX_SYSTEM
-static char *expand_link_1(char *f) /*{{{*/
+static char *expand_link_2(char *f) /*{{{*/
 {
    char work[JED_MAX_PATH_LEN], lnk[JED_MAX_PATH_LEN];
-   char *d = jed_expand_filename (f);
+   char *d = jed_standardize_filename (f);
    unsigned int max_num = 20;
 
    while ((d != NULL) && is_link(d, lnk))
@@ -1365,7 +1378,7 @@ static char *expand_link_1(char *f) /*{{{*/
 	     safe_strcat (work, lnk, sizeof (work));
 	  }
 
-	/* Collapse multiple // since expand_filename will interpret them
+	/* Collapse multiple // since jed_standardize_filename_static will interpret them
 	 * differently.
 	 */
 	w = w1 = work;
@@ -1383,28 +1396,17 @@ static char *expand_link_1(char *f) /*{{{*/
 	     jed_verror ("possible circular-link detected");
 	     return NULL;
 	  }
-	d = jed_expand_filename(work);
+	d = jed_standardize_filename(work);
      }
    
    return (d);
 }
 #endif				       /* REAL_UNIX_SYSTEM */
 
-/* This routine should be modified to work with all components of the path
- * and not just the last component.  For example, suppose /usr/lib/jed is a 
- * link to /usr/local/src/jed and one tries to read the file
- * /usr/lib/jed/lib/site.sl.  This should be expanded to 
- * /usr/local/lib/jed/lib/site.sl.  Currently, it will only have an effect if
- * site.sl is itself a link.
- * 
- * It returns a malloced string.
- */
-char *jed_expand_link(char *f) /*{{{*/
+#ifdef REAL_UNIX_SYSTEM
+static char *expand_link_1(char *f) /*{{{*/
 {
-#ifndef REAL_UNIX_SYSTEM
-   return jed_expand_filename (f);
-#else
-   char *d = jed_expand_filename (f);
+   char *d = SLmake_string (f);
    
    if (d == NULL)
      return d;
@@ -1426,7 +1428,7 @@ char *jed_expand_link(char *f) /*{{{*/
 	
 	*f++ = 0;
 
-	new_d = expand_link_1 (d);
+	new_d = expand_link_2 (d);
 	if (new_d == NULL)
 	  {
 	     SLfree (d);
@@ -1443,12 +1445,92 @@ char *jed_expand_link(char *f) /*{{{*/
      }
    
    /* Now get filename component */
-   f = expand_link_1 (d);
+   f = expand_link_2 (d);
    SLfree (d);
    return f;
+}
+#endif				       /* REAL_UNIX_SYSTEM */
+
+/* Malloced string returned */
+char *jed_expand_link (char *f)
+{
+#ifndef REAL_UNIX_SYSTEM
+   return jed_standardize_filename (f);
+#else
+   char *f1, *f2;
+   unsigned int len;
+   
+   /* Look for, e.g., /usr/local//tmp/foo and extract /tmp/foo.
+    * Do this before attempting to expand a link since links could 
+    * contain such constructs where they have a different meaning.
+    * 
+    * Do this only if // occurs because ../ constructs should be preserved
+    * if a subdir is a link to another.  That is consider:
+    * 
+    *    /path/a/b/../c.txt
+    * 
+    * jed_standardize_filename will convert this to
+    *  
+    *    /path/a/c.txt
+    * 
+    * However, if b is a symlink to /tmp/a/b then we really want the file
+    * /tmp/a/c.txt.
+    */
+#if 1   
+   len = strlen (f);
+   if (NULL == (f = SLmake_nstring (f, len)))
+     return NULL;
+   
+   f2 = f + len;
+   while (f2 > f)
+     {
+	if ((*f2-- == '/') && (*f2 == '/'))
+	  {
+	     f2++;
+	     break;
+	  }
+     }
+#else
+   if (NULL == (f = jed_standardize_filename (f)))
+     return NULL;
+   f2 = f;
+#endif
+
+   if (NULL == (f1 = expand_link_1 (f2)))
+     {
+	SLfree (f);
+	return NULL;
+     }
+
+   if (0 == strcmp (f, f1))
+     {
+	SLfree (f);
+	return f1;
+     }
+   SLfree (f);
+
+   f = f1;
+   /* Keep expanding the link until nothing changes */
+   while (1)
+     {
+	if (NULL == (f1 = expand_link_1 (f)))
+	  {
+	     SLfree (f);
+	     return NULL;
+	  }
+
+	if (0 == strcmp (f, f1))
+	  {
+	     SLfree (f1);
+	     return f;
+	  }
+	SLfree (f);
+	f = f1;
+     }
 #endif
 }
 
+
 /*}}}*/
 
 void visit_file (char *dir, char *file) /*{{{*/
@@ -1500,7 +1582,7 @@ char *jed_dir_file_merge(char *dir, char *file) /*{{{*/
    char *dirfile;
 
    if ((file == NULL) || (*file == 0))
-     return jed_expand_filename (dir);
+     return jed_standardize_filename (dir);
 
    
    dirfile = SLmalloc (strlen (dir) + strlen (file) + 2);
@@ -1510,7 +1592,7 @@ char *jed_dir_file_merge(char *dir, char *file) /*{{{*/
    fixup_dir (dirfile);
    strcat (dirfile, file);
 	
-   file = jed_expand_filename (dirfile);
+   file = jed_standardize_filename (dirfile);
    SLfree (dirfile);
    return file;
 }
@@ -1531,8 +1613,9 @@ int file_status(char *file) /*{{{*/
  *    -3 unknown error
  */
 {
-   int mode = 0;
-   short uid, gid;
+   mode_t mode = 0;
+   uid_t uid;
+   gid_t gid;
    return sys_chmod(file, 0, &mode, &uid, &gid);
 }
 
@@ -1540,12 +1623,38 @@ int file_status(char *file) /*{{{*/
 
 /* Returns 0 if unchanged, 1 if changed, or -1 if it does not exist
  * and buffer is unmodified.
+ * 
+ * What does it mean for the file to be changed on the disk if the buffer
+ * corresponds to a symlink?  That is, imagine editing foo.c which is a
+ * symlink to bar.c.  If bar.c changes, then foo.c shold be flagged as
+ * changed.  But, what if the link foo.c is changed such that it nolonger 
+ * points to bar.c?  Now the buffer foo.c is nolonger connected with 
+ * bar.c.
+ * 
+ * To handle both situations, do not use the dirfile field of the buffer
+ * structure, since that corresponds to bar.c and not the link foo.c.
+ * Since the file_changed_on_disk function uses stat and not lstat, if foo.c
+ * is a link to bar.c, then bar.c will be checked.
+ * 
+ * If the file has changed, be sure to update the buffers dirfile field and
+ * update the inode info.  If another buffer has the same inode info, then
+ * the user has shot himself in the foot by making foo.c a hardlink.
  */
-int file_changed_on_disk (Buffer *buf, char *file) /*{{{*/
+int file_changed_on_disk (Buffer *buf, char *dirfile) /*{{{*/
 {
    unsigned long t;
 
-   t = sys_file_mod_time(file);
+   /* If we are saving this to a different file, then skip the test */
+   if ((buf->canonical_dirfile != NULL) 
+#if JED_FILE_PRESERVE_CASE
+       && (0 != strcmp (buf->canonical_dirfile, dirfile))
+#else
+       && (0 != jed_case_strcmp (buf->canonical_dirfile, dirfile))
+#endif
+       )
+     return 0;
+
+   t = sys_file_mod_time(dirfile);
    if (t == 0)
      {
 	/* File does not exist.  If the buffer has not been modified since
@@ -1561,7 +1670,7 @@ int file_changed_on_disk (Buffer *buf, char *file) /*{{{*/
 
    if (t > buf->c_time)
      {
-	if (buf->c_time == 0)	       /* new buffer, writing to exiting file */
+	if (buf->c_time == 0)	       /* new buffer, writing to existing file */
 	  return 0;
 
 	return 1;
@@ -1609,7 +1718,7 @@ void check_buffer(Buffer *b) /*{{{*/
    /* Update (or even invalidate) inode information since directories may have
     * been renamed, etc...
     */
-   (void) get_inode_info (b->dir, &b->device, &b->inode);
+   (void) jed_get_inode_info (b->canonical_dirfile, &b->device, &b->inode);
 #endif
 
    b->flags &= ~FILE_MODIFIED;
@@ -1647,36 +1756,37 @@ void set_file_trans(int *mode) /*{{{*/
 
 int read_file_pointer(int fp) /*{{{*/
 {
-   int n = 1;
+   int n = 0;
    unsigned int num;
    unsigned char *vbuf;
    VFILE *vp;
+
+   if (SLang_get_error () || (vp = vstream(fp, MAX_LINE_LEN, VFile_Mode)) == NULL) return(-1);
    
-   if (SLang_Error || (vp = vstream(fp, MAX_LINE_LEN, VFile_Mode)) == NULL) return(-1);
-   
-   if (NULL == (vbuf = (unsigned char *) vgets(vp, &num))) return(0);
-   
-#ifdef KEEP_SPACE_INFO
-   if (CLine->space < num)
-#endif
-     remake_line(CLine->len + num + 1);
-   
-   SLMEMCPY((char *) CLine->data, (char *) vbuf, (int) num);
-   CLine->len = num;
-   
-   while(NULL != (vbuf = (unsigned char *) vgets(vp, &num)))
+   if (NULL != (vbuf = (unsigned char *) vgets(vp, &num)))
      {
 	n++;
-	if ((num == 1) && (*vbuf == '\n')) make_line(num); else make_line(num + 1);
+#ifdef KEEP_SPACE_INFO
+	if (CLine->space < num)
+#endif
+	  remake_line(CLine->len + num + 1);
+	
 	SLMEMCPY((char *) CLine->data, (char *) vbuf, (int) num);
 	CLine->len = num;
-	if (SLang_Error || SLKeyBoard_Quit) break;
+   
+	while(NULL != (vbuf = (unsigned char *) vgets(vp, &num)))
+	  {
+	     n++;
+	     if ((num == 1) && (*vbuf == '\n')) make_line(num); else make_line(num + 1);
+	     SLMEMCPY((char *) CLine->data, (char *) vbuf, (int) num);
+	     CLine->len = num;
+	     if (SLang_get_error () || SLKeyBoard_Quit) break;
+	  }
+	if (vp->cr_flag) CBuf->flags |= ADD_CR_ON_WRITE_FLAG;
+	else CBuf->flags &= ~ADD_CR_ON_WRITE_FLAG;
      }
+   
    if (vp->buf != NULL) SLfree(vp->buf);
-   
-   if (vp->cr_flag) CBuf->flags |= ADD_CR_ON_WRITE_FLAG;
-   else CBuf->flags &= ~ADD_CR_ON_WRITE_FLAG;
-   
    SLfree ((char *)vp);
    return n;
 }
@@ -1684,30 +1794,21 @@ int read_file_pointer(int fp) /*{{{*/
 /*}}}*/
 
 #ifdef REAL_UNIX_SYSTEM
-int get_inode_info (char *dir, int *device, int *inode) /*{{{*/
+
+int jed_get_inode_info (char *dirfile, int *device, int *inode)
 {
    struct stat st;
-   char work [JED_MAX_PATH_LEN + 512];
-   char *f;
-   
-   f = extract_file (dir);
-   if (*f != 0)
-     {
-	unsigned int len = f - dir;
-	strncpy (work, dir, len);
-	work[len] = 0;
-	dir = work;
-     }
    
    *inode = *device = -1;
-   if (-1 == stat (dir, &st)) return -1;
+   if ((dirfile == NULL)
+       || (-1 == stat (dirfile, &st)))
+     return -1;
    
    *inode = st.st_ino;
    *device = st.st_dev;
-   
+
    return 0;
 }
-/*}}}*/
 
 #endif
 
@@ -1729,7 +1830,7 @@ void jed_set_umask (int mask)
 #ifdef REAL_UNIX_SYSTEM
 	int u;
 	
-	u = umask (mask & 0777);
+	u = umask (mask & 07777);
 	if (default_umask == 0) default_umask = u;
 #endif
      }
@@ -1763,11 +1864,6 @@ int jed_dirfile_to_dir_file (char *dirfile, char **dirp, char **filep)
 }
 
 
-char *jed_expand_filename (char *file)
-{
-   return SLmake_string (expand_filename (file));
-}
-
 char *jed_get_canonical_pathname (char *file)
 {
    return jed_expand_link (file);
