@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -57,7 +57,7 @@
 #endif
 
 #ifdef REALLY_HAVE_TERMIOS_H
-# ifndef __os2__
+# if !defined(__os2__) && !defined(__APPLE__)
 #  define USE_PTY
 # endif
 #endif
@@ -68,6 +68,7 @@
 #include "misc.h"
 #include "jprocess.h"
 #include "paste.h"
+#include "sysdep.h"
 #include "sig.h"
 #include "cmds.h"
 
@@ -307,14 +308,15 @@ static void get_process_status (Process_Type *p) /*{{{*/
 #endif
 
    memset ((char *)p, 0, sizeof (Process_Type));
-   
+   p->rd = p->wd = -1;
+
    /* Adjust the array of read descriptors */
 
    i = 0;
    
    while (i < Num_Subprocesses)
      {
-	if (Subprocess_Read_fds[i][0] == slfd)
+	if (Subprocess_Read_fds[i][1] == slfd)
 	  break;
 	  
 	i++;
@@ -385,6 +387,16 @@ void jed_get_child_status (void) /*{{{*/
 {
    Process_Type *p, *pmax;
    
+   Ignore_User_Abort++;
+   if (SLang_get_error ())
+     {
+	Ignore_User_Abort--;
+	return;
+     }
+
+   /* FIXME: We should block SIGCHLD before trying to access this flag,
+    * or anything in this function for that matter.
+    */
    Child_Status_Changed_Flag--;
    
    get_process_input (&Number_Zero);
@@ -401,6 +413,9 @@ void jed_get_child_status (void) /*{{{*/
 	  }
 	p++;
      }
+
+   Ignore_User_Abort--;
+
 }
 
 /*}}}*/
@@ -474,40 +489,46 @@ static void child_signal_handler (int sig) /*{{{*/
 
 static int get_master_slave_fds (int *slave_read, int *slave_write,
 				 int *master_read, int *master_write,
-				 char *slave_tty_name, int *is_pty)
+				 char *slave_tty_name, int *is_pty,
+				 int want_pty)
 {
 #ifdef USE_PTY
-   int master;
-   
-   if (-1 == pty_open_master_pty (&master, slave_tty_name))
-     return -1;
-
-   *master_read = *master_write = master;
-   *slave_read = *slave_write = -1;
-   
-   *is_pty = 1;
-   return 0;
-#else
-
-   int fds0[2], fds1[2];
-   
-   if (-1 == pipe (fds0))
-     return -1;
-   
-   if (-1 == pipe (fds1))
+   if (want_pty)
      {
-	signal_safe_close (fds0[0]);
-	signal_safe_close (fds0[1]);
-	return -1;
+	int master;
+	
+	if (-1 == pty_open_master_pty (&master, slave_tty_name))
+	  return -1;
+
+	*master_read = *master_write = master;
+	*slave_read = *slave_write = -1;
+	
+	*is_pty = 1;
      }
-   *master_read = fds0[0];
-   *slave_write = fds0[1];
-   *master_write = fds1[1];
-   *slave_read = fds1[0];
-   *slave_tty_name = 0;
-   *is_pty = 0;
-   return 0;
+   else
+     {
 #endif
+	int fds0[2], fds1[2];
+   
+	if (-1 == pipe (fds0))
+	  return -1;
+	
+	if (-1 == pipe (fds1))
+	  {
+	     signal_safe_close (fds0[0]);
+	     signal_safe_close (fds0[1]);
+	     return -1;
+	  }
+	*master_read = fds0[0];
+	*slave_write = fds0[1];
+	*master_write = fds1[1];
+	*slave_read = fds1[0];
+	*slave_tty_name = 0;
+	*is_pty = 0;
+#ifdef USE_PTY
+     }
+#endif
+   return 0;
 }
 
 
@@ -518,7 +539,7 @@ static void my_setenv (char *what, char *value)
 #else
 # ifdef HAVE_PUTENV
    char buf[512];
-   _SLsnprintf (buf, sizeof (buf), "%s=%s", what, value);
+   SLsnprintf (buf, sizeof (buf), "%s=%s", what, value);
    (void) putenv (buf);
 # endif
 #endif
@@ -533,7 +554,7 @@ static void my_unsetenv (char *what)
 
    
 #ifdef __os2__
-static int open_process (char *pgm, char **argv) /*{{{*/
+static int open_process (char *pgm, char **argv, int want_pty) /*{{{*/
 {
     int val;
     int pd;
@@ -562,7 +583,7 @@ static int open_process (char *pgm, char **argv) /*{{{*/
    
    if (-1 == get_master_slave_fds (&slave_read, &slave_write,
 				   &master_read, &master_write,
-				   slave_tty_name, &p->is_pty))
+				   slave_tty_name, &p->is_pty, want_pty))
      {
 	SLang_free_mmt (mmt);
 	return -1;
@@ -733,7 +754,7 @@ static void sigtrap_handler (int sig)
 }
 #endif
    
-static int open_process (char *pgm, char **argv) /*{{{*/
+static int open_process (char *pgm, char **argv, int want_pty) /*{{{*/
 {
    int pd;
    int slave_read, slave_write, master_read, master_write;
@@ -759,7 +780,7 @@ static int open_process (char *pgm, char **argv) /*{{{*/
    
    if (-1 == get_master_slave_fds (&slave_read, &slave_write,
 				   &master_read, &master_write,
-				   slave_tty_name, &p->is_pty))
+				   slave_tty_name, &p->is_pty, want_pty))
      {
 	SLang_free_mmt (mmt);
 	return -1;
@@ -801,18 +822,6 @@ static int open_process (char *pgm, char **argv) /*{{{*/
 	  signal_safe_close (master_write);	       /* close write end of 0 */
 	signal_safe_close (master_read);	       /* close read end of 1 */
 
-	for (i = 0; i < 32; i++) 
-	  {
-#ifdef SIGTRAP
-	/* Under Linux, subprocesses fail if run under the debugger because
-	 * something sends this child a SIGTRAP.  Let's try ignoring it.
-	 */
-	     if (i == SIGTRAP)
-	       SLsignal (i, sigtrap_handler);
-	     else
-#endif
-	       SLsignal (i, SIG_DFL);
-	  }
 #ifdef USE_PTY	
 	/* Call set setsid so that the child will become the session leader.
 	 * This has the side effect that we will loose the controlling
@@ -851,6 +860,18 @@ static int open_process (char *pgm, char **argv) /*{{{*/
 	my_setenv ("TERM", "unknown");
 	my_unsetenv ("TERMCAP");
 
+	for (i = 0; i < 32; i++) 
+	  {
+#ifdef SIGTRAP
+	/* Under Linux, subprocesses fail if run under the debugger because
+	 * something sends this child a SIGTRAP.  Let's try ignoring it.
+	 */
+	     if (i == SIGTRAP)
+	       SLsignal (i, sigtrap_handler);
+	     else
+#endif
+	       SLsignal (i, SIG_DFL);
+	  }
 #ifdef SIGTRAP
 	SLsignal (SIGTRAP, SIG_DFL);
 #endif
@@ -900,6 +921,13 @@ static int open_process (char *pgm, char **argv) /*{{{*/
 /* This function is only called when we are reading characters from the 
  * keyboard.  Keyboard input has the highest priority and this is called only
  * if there is no input ready.
+ * 
+ * It might also get called by the hooks that it calls.  As a result, it must
+ * be reentrant.
+ * 
+ * It is possible for the user to press Ctrl-G while this routine is executing.
+ * This could cause slang hooks to fail.  It would probably be better to 
+ * block SIGINT for the duration of this routine.
  */
 void read_process_input (int fd) /*{{{*/
 {
@@ -915,6 +943,13 @@ void read_process_input (int fd) /*{{{*/
 	return;
      }
    
+   Ignore_User_Abort++;
+   if (SLang_get_error ())
+     {
+	Ignore_User_Abort--;
+	return;
+     }
+   
    otype = p->output_type;
    pbuf = p->buffer;
    
@@ -934,15 +969,19 @@ void read_process_input (int fd) /*{{{*/
 	if (otype & PROCESS_USE_BUFFER) 
 	  {
 	     if (0 == (otype & PROCESS_AT_POINT)) eob ();
-	     (void) jed_ins_chars (buf, n);
+	     (void) jed_insert_nbytes (buf, n);
 	     jed_move_user_object_mark (p->umark);
 	  }
 	else if (otype == PROCESS_USE_SLANG)
 	  {
 	     buf[n] = 0;
-	     SLang_push_integer ((int) (p - Processes));
-	     SLang_push_string ((char *) buf);
-	     SLexecute_function (p->slang_fun);    /* function to pass output to */
+	     if ((-1 == SLang_push_integer ((int) (p - Processes)))
+		 || (-1 == SLang_push_string ((char *) buf))
+		 || (-1 == SLexecute_function (p->slang_fun)))
+	       {
+		  /* It looks like the contents of buf may be lost */
+		  break;
+	       }
 	  }
      }
    
@@ -952,8 +991,14 @@ void read_process_input (int fd) /*{{{*/
    if (p->buffer != NULL)
      {
 	if (b != CBuf) switch_to_buffer (b);
-	pbuf->locked--;
      }
+   
+   /* Since it was locked, it cannot be deleted.  So pbuf is still a good 
+    * pointer
+    */
+   if (pbuf != NULL)
+     pbuf->locked--;
+
    if (total)
      {
 #ifdef __os2__
@@ -963,6 +1008,8 @@ void read_process_input (int fd) /*{{{*/
 #endif
 	touch_screen ();
      }
+   
+   Ignore_User_Abort--;
 }
 
 
@@ -970,17 +1017,47 @@ void read_process_input (int fd) /*{{{*/
 
 static int write_to_process (int fd, char *buf, unsigned int len)
 {
+   int n, ilen;
+   
    if (len == 0) 
      return 0;
    
-   /* FIXME!! check for write errors!! */
-   while (-1 == write (fd, buf, len))
+   ilen = (int) len;
+   n = 0;
+   while (n < ilen)
      {
-	if (errno != EINTR)
+	int dn;
+	int count = 0;
+
+	while (-1 == (dn =  write (fd, buf + n, ilen - n)))
+	  {
+	     if (errno == EINTR)
+	       continue;
+	     
+#ifdef EAGAIN
+	     if (errno == EAGAIN)
+	       {
+		  get_process_input (&Number_Zero);
+		  count++;
+		  if ((count % 10) == 0)
+		    {
+		       sleep (1);
+		       if (count > 20)
+			 return n;
+		    }
+		  continue;
+	       }
+#endif
+	     return -1;
+	  }
+
+	if (dn == 0)
 	  break;
+
+	n += dn;
      }
-   
-   return 0;
+  
+   return n;
 }
 
 int jed_send_process (int *fd, char *str) /*{{{*/
@@ -991,7 +1068,10 @@ int jed_send_process (int *fd, char *str) /*{{{*/
    
    len = strlen (str);
    
-   return write_to_process (p->wd, str, len);
+   if (len != (unsigned int) write_to_process (p->wd, str, len))
+     jed_verror ("write to process failed");
+   
+   return len;
 }
 
 /*}}}*/
@@ -1110,11 +1190,11 @@ void jed_get_process_mark (int *fd) /*{{{*/
 
 /*}}}*/
 
-int jed_open_process (int *np) /*{{{*/
+static int open_process_of_type (int nargs, int wantpty) /*{{{*/
 {
    int fd = -1;
    char *argv[502];
-   int n = *np;
+   int n = nargs;
 
    if (CBuf->subprocess)
      {
@@ -1141,8 +1221,8 @@ int jed_open_process (int *np) /*{{{*/
 	  }
      }
    n = 0;
-   
-   if ((fd = open_process(argv[0], argv)) < 0)
+
+   if ((fd = open_process(argv[0], argv, wantpty)) < 0)
      {
 	jed_verror ("Unable to open %s process.", argv[0]);
      }
@@ -1150,7 +1230,7 @@ int jed_open_process (int *np) /*{{{*/
    /* free up the argument strings */
    free_return:
    
-   while (n <= *np)
+   while (n <= nargs)
      {
 	SLang_free_slstring (argv[n]);
 	n++;
@@ -1161,6 +1241,17 @@ int jed_open_process (int *np) /*{{{*/
 
 /*}}}*/
 
+int jed_open_process (int *np)
+{
+   return open_process_of_type (*np, 1);
+}
+
+int jed_open_process_pipe (int *np)
+{
+   return open_process_of_type (*np, 0);
+}
+
+
 void jed_block_child_signal (int block) /*{{{*/
 {
    static sigset_t new_mask, old_mask;
