@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -50,20 +50,12 @@ static void touch_windows(void) /*{{{*/
 	     n = CBuf->marks->n;
 	     while ((l != NULL) && (l != JWindow->mark.line)) l = l->next, n++;
 	     if (l == NULL) 
-	       {
-#if 0
-		  JWindow->mark.line = CLine;
-		  JWindow->mark.point = Point;
-		  JWindow->mark.n = n - 1;
-#else
-		  jed_init_mark (&JWindow->mark, 0);
-#endif
-	       }
+	       jed_init_mark (&JWindow->mark, 0);
 	     touch_window();
 	  }
 	JWindow = JWindow->next;
      }
-   pop_mark(&Number_Zero);
+   jed_pop_mark(0);
 }
 
 /*}}}*/
@@ -157,14 +149,14 @@ int set_mark_cmd (void) /*{{{*/
    Mark *m;
    if (Repeat_Factor != NULL)
      {
-	while (CBuf->marks != NULL) pop_mark(&Number_Zero);
+	while (CBuf->marks != NULL) jed_pop_mark(0);
 	Repeat_Factor = NULL;
 	return(1);
      }
    
    if (CBuf->marks == NULL)
      {
-	push_mark ();
+	jed_push_mark ();
      }
 
    m = CBuf->marks;
@@ -242,7 +234,7 @@ int push_spot() /*{{{*/
 
 /*}}}*/
 
-int push_mark() /*{{{*/
+int jed_push_mark() /*{{{*/
 {
    (void) allocate_mark (&CBuf->mark_array, &CBuf->marks, 0);
    return 1;
@@ -272,14 +264,14 @@ int jed_goto_mark(Mark *m) /*{{{*/
 
 /*}}}*/
 
-int pop_mark (int *go) /*{{{*/
+int jed_pop_mark (int go) /*{{{*/
 {
    Mark *m;
    
    m = CBuf->marks;
    if (m == NULL) return(0);
 
-   if (*go) jed_goto_mark(m);
+   if (go) jed_goto_mark(m);
    if (m->flags & VISIBLE_MARK_MASK)
      {
 	CBuf->vis_marks--;
@@ -400,13 +392,13 @@ static int widen_buffer_lines (Buffer *b) /*{{{*/
    push_spot();
    eob();
    if ((n->end != NULL)
-       && (!CLine->len || ('\n' != *(CLine->data + (CLine->len - 1)))))
+       && (0 == LINE_HAS_NEWLINE (CLine)))
      {
 	/* This is a hack to avoid messing with the lock file. Yuk. */
 	unsigned int flags = CBuf->flags;
 	CBuf->flags &= ~(READ_ONLY);
 	CBuf->flags |= BUFFER_NON_LOCKING;
-	(void) jed_ins('\n');
+	(void) _jed_ins_byte ('\n');
 	CBuf->flags = flags;  /* jed_set_buffer_flags (CBuf, flags); */
      }
 
@@ -467,18 +459,18 @@ int widen_buffer (Buffer *b) /*{{{*/
    line_flags = CLine->flags;
    CLine->flags &= ~JED_LINE_IS_READONLY;
 #endif
-   (void) jed_del ();
+   (void) jed_del_wchar ();
 #if JED_HAS_LINE_ATTRIBUTES
    CLine->flags = line_flags;
 #endif
    
    pop_spot ();
-   prevline (&Number_One);
+   (void) jed_up (1);
 #if JED_HAS_LINE_ATTRIBUTES
    line_flags = CLine->flags;
    CLine->flags &= ~JED_LINE_IS_READONLY;
 #endif
-   (void) jed_del ();
+   (void) jed_del_wchar ();
 #if JED_HAS_LINE_ATTRIBUTES
    CLine->flags = line_flags;
 #endif
@@ -517,7 +509,7 @@ int narrow_to_region (void) /*{{{*/
    line = CLine;
    pnt = Point;
    
-   pop_mark (&Number_One);
+   jed_pop_mark (1);
    /* Special case if region is empty */
    if ((CLine == line) && (pnt == Point))
      {
@@ -535,14 +527,14 @@ int narrow_to_region (void) /*{{{*/
 	if (CLine->prev != NULL) CLine->prev->flags = line_flags;
 #endif
 	
-	push_mark ();
+	jed_push_mark ();
 
 	(void) jed_insert_newline ();
 	
 #if JED_HAS_LINE_ATTRIBUTES
 	CLine->flags = line_flags;
 #endif
-	prevline (&Number_One);
+	(void) jed_up (1);
 
 #if JED_HAS_LINE_ATTRIBUTES
 	CLine->flags = line_flags;
@@ -563,7 +555,7 @@ int narrow_to_region (void) /*{{{*/
 	if (CLine->prev != NULL) CLine->prev->flags = line_flags;
 #endif
 
-	push_mark ();
+	jed_push_mark ();
 	pop_spot ();
 	
 #if JED_HAS_LINE_ATTRIBUTES
@@ -574,7 +566,7 @@ int narrow_to_region (void) /*{{{*/
 #if JED_HAS_LINE_ATTRIBUTES
 	CLine->flags = line_flags;
 #endif
-	prevline (&Number_One);
+	jed_up (1);
 
 #if JED_HAS_LINE_ATTRIBUTES
 	CLine->flags = line_flags;
@@ -594,16 +586,20 @@ int narrow_to_region (void) /*{{{*/
 
 int widen (void) /*{{{*/
 {
-   /* unmark_undo_boundary (CBuf); */
-   return widen_buffer_lines (CBuf);
+   if (CBuf->narrow == NULL)
+     return 0;
+
+   if (CBuf->narrow->is_region == 0)
+     return widen_buffer_lines (CBuf);
+
+   return widen_buffer (CBuf);
 }
 
 /*}}}*/
 
 int widen_region (void) /*{{{*/
 {
-   /* unmark_undo_boundary (CBuf); */
-   return widen_buffer (CBuf);
+   return widen ();
 }
 
 /*}}}*/
@@ -622,8 +618,8 @@ int narrow_to_lines () /*{{{*/
    /* unmark_undo_boundary (CBuf); */
 
    push_spot();
-   pop_mark(&Number_One);
-   push_mark();			       /* popped and used in touch_windows! */
+   jed_pop_mark(1);
+   jed_push_mark();			       /* popped and used in touch_windows! */
    beg = CLine;
    nt->nup = LineNum - 1;
    
@@ -678,7 +674,7 @@ int yank() /*{{{*/
 
 int copy_region_to_buffer(Buffer *b) /*{{{*/
 {
-   int first_point, last_point, n, tmpm;
+   int first_point, last_point, n;
    Line *first, *last;
    Buffer *save_buf;
 
@@ -692,7 +688,7 @@ int copy_region_to_buffer(Buffer *b) /*{{{*/
    last = CLine;
    last_point = Point;
 
-   tmpm = 1; pop_mark(&tmpm);
+   jed_pop_mark(1);
    if (b == CBuf)
      {
 	msg_error("A buffer cannot be inserted upon itself.");
@@ -713,9 +709,9 @@ int copy_region_to_buffer(Buffer *b) /*{{{*/
 	n = last_point - first_point;
 	if (save_buf == MiniBuffer)
 	  {
-	     (void) jed_ins_chars(first->data + first_point, n);
+	     (void) jed_insert_nbytes (first->data + first_point, n);
 	  }
-	else (void) jed_quick_insert(first->data + first_point, n);
+	else (void) jed_quick_insert (first->data + first_point, n);
      }
    else 
      {
@@ -725,7 +721,7 @@ int copy_region_to_buffer(Buffer *b) /*{{{*/
 
 	while (first = first->next, first != last)
 	  {
-	     if (-1 == jed_quick_insert(first->data, first->len))
+	     if (-1 == jed_quick_insert (first->data, first->len))
 	       goto the_return;
 	  }
 	(void) jed_quick_insert(first->data, last_point);
@@ -751,72 +747,103 @@ int copy_to_pastebuffer() /*{{{*/
 
 /*}}}*/
 
+static int jed_check_readonly_region (void)
+{
+   int beg_point, end_point;
+   Line *beg, *end;
+
+   if (CBuf->flags & READ_ONLY) 
+     { 
+	msg_error(Read_Only_Error);
+	return -1;
+     }
+
+   if (!check_region(&Number_Zero))
+     return -1;
+
+   end = CLine; end_point = Point;
+   (void) exchange_point_mark ();
+   beg = CLine; beg_point = Point;
+   (void) exchange_point_mark ();
+   
+   while (1)
+     {
+	if (beg->flags & JED_LINE_IS_READONLY) 
+	  {
+	     if ((beg == end) 
+		 && (end_point == 0)
+		 && (beg_point == 0))
+	       return 0;
+
+	     msg_error (Line_Read_Only_Error);
+	     return -1;
+	  }
+	if (beg == end)
+	  break;
+
+	beg = beg->next;
+     }
+   return 0;
+}
+
+   
 int delete_region (void) /*{{{*/
 {
-   int beg_point, tmpm, end_point, n;
+   int beg_point, end_point;
    Line *beg, *end;
    
-   CHECK_READ_ONLY
-    if (!check_region(&Number_Zero)) return(0);
+   if (0 != jed_check_readonly_region ())
+     return -1;
 
    /* make this go through standard ins/del routines to ensure undo */
    
    end = CLine; end_point = Point;
    push_spot();
-   tmpm = 1; pop_mark(&tmpm);
+   jed_pop_mark(1);
    beg = CLine; beg_point = Point;
    pop_spot();
    
-   Point = 0;
-   
    if (end != beg) 
      {
-	if (-1 == jed_deln (end_point))
+	bol ();
+   
+	if (-1 == jed_del_nbytes (end_point))
 	  return -1;
 
 	/* go back because we do not want to mess with Line structures
 	   changing on us --- shouldn't happen anyway */
 	
-	while (CLine = CLine->prev, LineNum--, CLine != beg)
+	while (jed_up (1) && (CLine != beg))
 	  {
-	     eol(); 
-	     n = Point + 1;
-	     Point = 0;
-	     if (-1 == jed_generic_deln (n))
+	     bol ();
+	     if (-1 == jed_del_through_eol ())
 	       return -1;
 	  }
-	eol();
-	n = Point - beg_point + 1;  /* the \n char */
-	Point = beg_point;
+	end_point = CLine->len;	       /* include \n */
      }
-   else 
-     {
-	Point = beg_point;
-	n = end_point - Point;
-     }
-   
-   (void) jed_generic_deln (n);
-   return(1);
+
+   jed_set_point (beg_point);
+   (void) jed_generic_del_nbytes (end_point - Point);
+   return 1;
 }
 
 /*}}}*/
 
 int kill_region() /*{{{*/
 {
-   int tmpm = 1;
-   
-   CHECK_READ_ONLY
+   if (-1 == jed_check_readonly_region ())
+     return -1;
 
    /* need two marks for this one */
    push_spot();
-   if (!pop_mark(&tmpm))
+   if (!jed_pop_mark(1))
      {
 	check_region(&Number_Zero);
 	pop_spot();
 	return(0);
      }
-   push_mark();
-   push_mark();
+   jed_push_mark();
+   jed_push_mark();
    pop_spot();
 
    copy_to_pastebuffer();
@@ -837,7 +864,11 @@ int insert_rectangle() /*{{{*/
    Line *rline;
 
    CHECK_READ_ONLY
-   if (Rectangle_Buffer == NULL) return(0);
+   if (0 == buffer_exists (Rectangle_Buffer))
+     {
+	Rectangle_Buffer = NULL;
+	return 0;
+     }
 
    Suspend_Screen_Update = 1;
    c1 = calculate_column();
@@ -850,16 +881,11 @@ int insert_rectangle() /*{{{*/
 
 	rline = rline->next;
 	if (rline == NULL) break;
-	if (CLine->next == NULL)
+	if (0 == jed_down (1))
 	  {
 	     eol();
 	     jed_insert_newline();
 	  }
-	else 
-	  {
-	     CLine = CLine->next;
-	     LineNum++;
-	  }
      }
    return(1);
 }
@@ -876,7 +902,7 @@ int open_rectangle() /*{{{*/
 
    c1 = calculate_column();
    save_line = CLine;
-   tmpm = 1; pop_mark(&tmpm);
+   tmpm = 1; jed_pop_mark(1);
    c2 = calculate_column();
    n = c2 - c1;
    if (n < 0)
@@ -889,11 +915,10 @@ int open_rectangle() /*{{{*/
    while(1)
      {
 	goto_column(&c1);
-	if (-1 == jed_ins_char_n_times(' ', n))
+	if (-1 == jed_insert_wchar_n_times(' ', n))
 	  break;
 	if (CLine == save_line) break;
-	CLine = CLine->next;
-	LineNum++;
+	(void) jed_down (1);
      }
    pop_spot();
 
@@ -902,151 +927,99 @@ int open_rectangle() /*{{{*/
 
 /*}}}*/
 
-/* rectangle commands */
-int copy_rectangle() /*{{{*/
+/* MB Safe */
+static int copy_or_kill_rectangle (int kill) /*{{{*/
 {
-    Line *save_line, *line, *beg;
-    int c1, c2, dc, tmp, tmpm, dc_malloc;
-    unsigned char *p1, *p2, *data;
+   Line *save_line, *line, *beg;
+   int c1, c2, dc, tmp;
    
-    if (!check_region(&Number_One)) return(0);       /* spot pushed */
-    /* delete Rectangle buffer */
-    if (Rectangle_Buffer != NULL) delete_buffer(Rectangle_Buffer);
-
-    Rectangle_Buffer = make_buffer (" <rect>", NULL, NULL);
-    c2 = calculate_column();
-    save_line = CLine;
-
-    tmpm = 1; pop_mark(&tmpm);
-    c1 = calculate_column();
-    if (c1 == c2)
-      {
-	 msg_error(Rect_Error);
-	 pop_spot();
-	 return(0);
-      }
-    if (c1 > c2)
-      {
-	 tmp = c1;
-	 c1 = c2;
-	 c2 = tmp;
-	 goto_column(&c1);
-      }
-
-    /* go through the region copying rectanglar blocks to Rectanglebuffer */
-    dc = c2 - c1;
-   if (dc == 1) dc_malloc = 2; else dc_malloc = dc;
-    line = beg = make_line1(dc_malloc);
-   
-    beg->prev = NULL;
-    while (1)
-      {
-	 data = line->data;
-	 /* p1 = data;
-	 p2 = data + dc;
-	 while (p1 < p2) *p1++ = ' '; */
-	 SLMEMSET ((char *) data, ' ', dc);
-	 
-	 line->len = dc;
-
-	 if (c1 == goto_column1(&c1))
-	   {
-	      p1 = CLine->data + Point;
-	      (void) goto_column1(&c2);
-	      p2 = CLine->data + Point;
-
-	      /* while(p1 < p2) *data++ = *p1++; */
-	      SLMEMCPY((char *) data, (char *) p1, (int) (p2 - p1));
-	   }
-	 if (CLine == save_line) break;
-	 CLine = CLine->next;
-	 LineNum++;
-
-	 line->next = make_line1(dc_malloc);
-	 line->next->prev = line;
-	 line = line->next;
-      }
-
-    line->next = NULL;
-
-    Rectangle_Buffer->line = Rectangle_Buffer->beg = beg;
-    Rectangle_Buffer->end = line;
-    Rectangle_Buffer->point = 0;
-
-    pop_spot();
-    return(0);
-}
-
-/*}}}*/
-
-int kill_rectangle() /*{{{*/
-{
-    Line *save_line, *line, *beg;
-   int c1, c2, dc, tmp, n, tmpm, dc_malloc;
-   unsigned char *p1, *p2, *data;
-
-   CHECK_READ_ONLY
-   if (!check_region(&Number_One)) return(0);
-
+   if (!check_region(&Number_One)) return(0);       /* spot pushed */
    /* delete Rectangle buffer */
    if (Rectangle_Buffer != NULL) delete_buffer(Rectangle_Buffer);
-
+   
    Rectangle_Buffer = make_buffer (" <rect>", NULL, NULL);
    c2 = calculate_column();
    save_line = CLine;
+   
+   jed_pop_mark(1);
+   c1 = calculate_column();
+   if (c1 == c2)
+     {
+	msg_error(Rect_Error);
+	pop_spot();
+	return(0);
+     }
+   if (c1 > c2)
+     {
+	tmp = c1;
+	c1 = c2;
+	c2 = tmp;
+	goto_column(&c1);
+     }
+   
+   dc = c2 - c1;
+   
+   /* go through the region copying rectanglar blocks to Rectanglebuffer */
 
-   tmpm = 1; pop_mark(&tmpm);
-    c1 = calculate_column();
-    if (c1 == c2)
-      {
-	 msg_error(Rect_Error);
-	 pop_spot();
-	 return(0);
-      }
-    if (c1 > c2)
-      {
-	 tmp = c1;
-	 c1 = c2;
-	 c2 = tmp;
-	 goto_column(&c1);
-      }
+   line = beg = NULL;
+   while (1)
+     {
+	int col2, col1;
+	unsigned int len, len1;
+	unsigned int nspaces;
+	unsigned int nbytes;
+	unsigned char *p1, *p2;
 
-   Suspend_Screen_Update = 1;
-    /* go through the region copying rectanglar blocks to Rectanglebuffer */
-    dc = c2 - c1;
-   if (dc == 1) dc_malloc = 2; else dc_malloc = dc;
-   		       /* length of 1 not usable here.
-			* See makeline1 to see why 
-			*/
-    line = beg = make_line1(dc_malloc);
-    beg->prev = NULL;
-    while (1)
-      {
-	 data = line->data;
-	 
-	 /* p2 = data + dc; p1 = data;
-	 while (p1 < p2) *p1++ = ' '; */
-	 SLMEMSET((char *) data, ' ', dc);
-	 line->len = dc;
+	col1 = goto_column1 (&c1);
+	p1 = CLine->data + Point;
 
-	 if (c1 == goto_column1(&c1))
-	   {
-	      p1 = CLine->data + Point;
-	      (void) goto_column1(&c2);
-	      p2 = CLine->data + Point;
-	      Point = (int) (p1 - CLine->data);
-	      n = (int) (p2 - p1);
-	      SLMEMCPY((char *) data, (char *) p1, n);
+	if (col1 == c1)
+	  {
+	     col2 = goto_column1(&c2);
+	     p2 = CLine->data + Point;
+	  }
+	else
+	  {
+	     col2 = col1;
+	     p2 = p1;
+	  }
+	
+	nspaces = (unsigned int) (dc - (col2 - col1));
+	nbytes = (unsigned int) (p2 - p1);
+	len1 = len = nbytes + nspaces;
+	
+	/* Need to allocate at least 2 bytes, since a single byte is a 
+	 * signature of a line with a single character whose value is a 
+	 * newline character.  See make_line1 for details.
+	 */
+	if (len <= 1)
+	  len1++;
+
+	if (beg == NULL)
+	  {
+	     beg = line = make_line1 (len1);
+	     beg->prev = NULL;
+	  }
+	else
+	  {
+	     line->next = make_line1 (len1);
+	     line->next->prev = line;
+	     line = line->next;
+	  }
+	line->len = len;
+
+	memcpy ((char *) line->data, (char *) p1, nbytes);
+	memset ((char *) line->data + nbytes, ' ', nspaces);
+
+	if (kill && (nbytes != 0))
+	  {
+	     jed_position_point (p1);
+	     if (-1 == jed_del_nbytes (nbytes))
+	       break;
+	  }
 
-	      if (-1 == jed_deln (n))
-		break;
-	   }
 	 if (CLine == save_line) break;
-	 CLine = CLine->next;  LineNum++;
-
-	 line->next = make_line1(dc_malloc);
-	 line->next->prev = line;
-	 line = line->next;
+	 (void) jed_down(1);
       }
 
     line->next = NULL;
@@ -1061,49 +1034,56 @@ int kill_rectangle() /*{{{*/
 
 /*}}}*/
 
-int blank_rectangle() /*{{{*/
+int copy_rectangle (void)
 {
-   int c1, n, c2, pnt, tmpm;
+   return copy_or_kill_rectangle (0);
+}
+
+int kill_rectangle (void)
+{
+   return copy_or_kill_rectangle (1);
+}
+
+int blank_rectangle (void) /*{{{*/
+{
+   int c1, c2;
    Line *save_line;
-   int nn;
-   
+   int ncols;
 
    CHECK_READ_ONLY
    if (!check_region(&Number_One)) return(0); /* push_spot(); performed */
 
    c1 = calculate_column();
    save_line = CLine;
-   tmpm = 1; pop_mark(&tmpm);
+   jed_pop_mark(1);
    c2 = calculate_column();
-   n = c2 - c1;
-   if (n < 0)
+   if (c1 > c2)
      {
-	n = -n;
+	int tmp = c1;
 	c1 = c2;
+	c2 = tmp;
      }
-   
+   ncols = c2 - c1;
+
    Suspend_Screen_Update = 1;
    while(1)
      {
-	goto_column(&c1);
+	int pnt;
+
+	goto_column (&c2);
 	pnt = Point;
-	eol();
-	nn = Point - pnt;
-	if (nn > n) nn = n;
-	Point = pnt;
+	goto_column (&c1);
 	
-	if (-1 == jed_deln (nn))
+	if (-1 == jed_del_nbytes (pnt - Point))
 	  break;
 
-	if (-1 == jed_ins_char_n_times( ' ', nn))
+	if (-1 == jed_insert_wchar_n_times( ' ', ncols))
 	  break;
-	
+
 	if (CLine == save_line) break;
-	CLine = CLine->next;
-	LineNum++;
+	(void) jed_down (1);
      }
    pop_spot();
-   /* mark_buffer_modified(&Number_One); */
    return(1);
 }
 
@@ -1122,7 +1102,7 @@ typedef struct /*{{{*/
 /*}}}*/
 User_Mark_Type;
 
-static void free_user_mark (unsigned char type, VOID_STAR um_alias) /*{{{*/
+static void free_user_mark (SLtype type, VOID_STAR um_alias) /*{{{*/
 {
    Mark *m, *m1;
    Buffer *b;
@@ -1342,8 +1322,8 @@ char *user_mark_buffer (void) /*{{{*/
 /*}}}*/
 
 static int
-user_mark_bin_op_result (int op, unsigned char a, unsigned char b,
-			 unsigned char *c)
+user_mark_bin_op_result (int op, SLtype a, SLtype b,
+			 SLtype *c)
 {
    (void) a; (void) b;
    switch (op)
@@ -1366,8 +1346,8 @@ user_mark_bin_op_result (int op, unsigned char a, unsigned char b,
 
 static int
 user_mark_bin_op (int op,
-		  unsigned char a_type, VOID_STAR ap, unsigned int na,
-		  unsigned char b_type, VOID_STAR bp, unsigned int nb,
+		  SLtype a_type, VOID_STAR ap, unsigned int na,
+		  SLtype b_type, VOID_STAR bp, unsigned int nb,
 		  VOID_STAR cp)
 {
    int *ic;
@@ -1488,7 +1468,7 @@ void jed_create_line_mark (int *color) /*{{{*/
 /*}}}*/
 #endif
    
-static int user_mark_sget (unsigned char type, char *name)
+static int user_mark_sget (SLtype type, char *name)
 {
    SLang_MMT_Type *mmt;
    User_Mark_Type *um;
@@ -1562,7 +1542,7 @@ static void restore_saved_narrow (void) /*{{{*/
    while (beg != NULL)
      {
 	jed_goto_mark (beg);
-	push_mark ();
+	jed_push_mark ();
 	jed_goto_mark (end);
 	if (end->flags & NARROW_REGION_MARK)
 	  narrow_to_region ();
