@@ -1,4 +1,4 @@
-/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -6,8 +6,8 @@
  */
 #include "config.h"
 
-#define RGREP_MAJOR_VERSION 1
-#define RGREP_MINOR_VERSION 6
+#define RGREP_MAJOR_VERSION 2
+#define RGREP_MINOR_VERSION 1
 
 #include <stdio.h>
 #include <slang.h>
@@ -69,6 +69,25 @@ static int Stdout_Is_TTY;
 #define HOFF_STR "\033[0m"
 #define HOFF_STR_LEN 4
 
+#if SLANG_VERSION < 10410
+int SLang_get_error (void)
+{
+   return SLang_Error;
+}
+
+int SLang_set_error (int e)
+{
+   SLang_Error = e;
+   return 0;
+}
+#endif
+
+extern int jed_handle_interrupt (void);
+int jed_handle_interrupt (void)
+{
+   return 0;
+}
+
 static void version (void)
 {
    fprintf (stdout, "rgrep version %d.%d\n",
@@ -170,7 +189,6 @@ static void additional_help (void)
 
 static FILE *File_Fp;
 static VFILE *File_Vp;
-SLsearch_Type Search_St;
 static unsigned char *Fixed_Len_Buf;
 static int Fixed_Len_Mode;
 static int Fixed_Line_Len;
@@ -242,10 +260,15 @@ static void parse_flags(char *f)
 }
 
 
-
+#if SLANG_VERSION < 20000
 static SLRegexp_Type reg;
 static SLRegexp_Type recurse_reg;
 static unsigned char Recurse_Reg_Pattern_Buffer[JED_MAX_PATH_LEN];
+#else
+static SLRegexp_Type *reg;
+static SLRegexp_Type *recurse_reg;
+static unsigned char Recurse_Reg_Pattern_Buffer[JED_MAX_PATH_LEN];
+#endif
 static int Must_Match;
 static int print_file_too;
 
@@ -318,13 +341,24 @@ static unsigned char *rgrep_gets (unsigned int *n)
 }
 
 
+static int OSearch_Ok = 0;
+static int Search_St_Key_Length = 0;
+#if SLANG_VERSION < 20000
+static SLsearch_Type Search_St;
+# define SEARCH_FORWARD(a,b) SLsearch ((a), (b), &Search_St)
+# define REGEXP_MATCH(r,b,n) SLang_regexp_match((unsigned char *)(b),(n), &r)
+#else
+static SLsearch_Type *Search_St;
+# define SEARCH_FORWARD(a,b) SLsearch_forward (Search_St, (a), (b))
+# define REGEXP_MATCH(r,b,n) (NULL != SLregexp_match(r,(char *) (b),(n)))
+#endif
 
 static void grep(char *file)
 {
    unsigned char *buf, *p, *pmax;
    unsigned int n;
    int line = 0, n_matches = 0;
-   
+
    if (NULL == (buf = (unsigned char *) rgrep_gets(&n)))
      return;
    
@@ -361,6 +395,7 @@ static void grep(char *file)
    do
      {
 	line++;
+#if SLANG_VERSION < 20000
 	if (reg.min_length > n)
 	  {
 	     if (Print_Non_Matching_Lines)
@@ -371,9 +406,10 @@ static void grep(char *file)
 	       }
 	     continue;
 	  }
-	
+#endif
 	if (Must_Match)
 	  {
+#if SLANG_VERSION < 20000
 	     if (Search_St.key_len > (int) n)
 	       {
 		  if (Print_Non_Matching_Lines)
@@ -384,8 +420,8 @@ static void grep(char *file)
 		    }
 		  continue;
 	       }
-	     
-	     if (NULL == (p = SLsearch (buf, buf + n, &Search_St)))
+#endif
+	     if (NULL == (p = SEARCH_FORWARD (buf, buf + n)))
 	       {
 		  if (Print_Non_Matching_Lines)
 		    {
@@ -396,15 +432,15 @@ static void grep(char *file)
 		  
 		  continue;
 	       }
-	     if (reg.osearch)
+	     if (OSearch_Ok)
 	       {
 		  if (Print_Non_Matching_Lines) continue;
-		  pmax = p + Search_St.key_len;
+		  pmax = p + Search_St_Key_Length;
 		  goto match_found;
 	       }
 	  }
 	
-	if (!SLang_regexp_match(buf, (int) n, &reg))
+	if (!REGEXP_MATCH (reg, buf, n))
 	  {
 	     if (Print_Non_Matching_Lines)
 	       {
@@ -417,9 +453,17 @@ static void grep(char *file)
 
 	if (Print_Non_Matching_Lines) continue;
 	
+#if SLANG_VERSION < 20000
 	p = buf + reg.beg_matches[0];
 	pmax = p + reg.end_matches[0];
-	
+#else
+	  {
+	     unsigned int ofs, len;
+	     (void) SLregexp_nth_match (reg, 0, &ofs, &len);
+	     p = buf + ofs;
+	     pmax = p + len;
+	  }
+#endif
 	match_found:
 	n_matches++;
 
@@ -796,8 +840,8 @@ static void grep_file(char *file, char *filename)
 		 )
 	       return;
 	  }
-	else
-	  if (!SLang_regexp_match((unsigned char *) filename, strlen(filename), &recurse_reg)) return;
+	else if (!REGEXP_MATCH(recurse_reg, filename, strlen(filename)))
+	  return;
      }
    
    File_Fp = NULL;
@@ -885,8 +929,13 @@ static unsigned char *fixup_filename (unsigned char *name)
 
 int main(int argc, char **argv)
 {
+#if SLANG_VERSION < 20000
    unsigned char buf[2 * JED_MAX_PATH_LEN];
    unsigned char recurse_buf[JED_MAX_PATH_LEN];
+#else
+   unsigned int flags;
+#endif
+   char *pattern;
 
 #ifdef __unix__
    Stdout_Is_TTY = isatty (fileno(stdout));
@@ -903,15 +952,24 @@ int main(int argc, char **argv)
 	     argv++;
 	     if (!argc) usage();
 	     
+#if SLANG_VERSION < 20000
 	     recurse_reg.pat = fixup_filename ((unsigned char *) *argv);
 	     recurse_reg.buf = recurse_buf;
 	     recurse_reg.buf_len = JED_MAX_PATH_LEN;
-#ifdef __MSDOS__
+# ifdef __MSDOS__
 	     recurse_reg.case_sensitive = 0;
-#else
+# else
 	     recurse_reg.case_sensitive = 1;
-#endif
+# endif
 	     if (SLang_regexp_compile (&recurse_reg)) exit_error("Error compiling pattern.");
+#else
+	     flags = 0;
+# ifdef __MSDOS__
+	     flags |= SLREGEXP_CASELESS;
+# endif
+	     if (NULL == (recurse_reg = SLregexp_compile ((char *)fixup_filename((unsigned char *) *argv), flags)))
+	       exit_error ("Error compiling pattern");
+#endif
 	     Do_Recursive = 1;
 	     Recursive_Match = 1;
 	  }
@@ -932,27 +990,50 @@ int main(int argc, char **argv)
    
    if (!argc) usage();
    
-   reg.pat = (unsigned char *) *argv;
+   pattern = *argv;
+
+#if SLANG_VERSION < 20000
+   reg.pat = (unsigned char *) pattern;
    reg.buf = buf;
    reg.buf_len = sizeof (buf);
    reg.case_sensitive = Case_Sensitive;
    
    if (SLang_regexp_compile (&reg)) exit_error("Error compiling pattern.");
+#else
+   if (NULL == (reg = SLregexp_compile (pattern, Case_Sensitive ? 0 : SLREGEXP_CASELESS)))
+     exit_error ("Error compiling pattern");
+#endif
    argc--; argv++;
    
    Must_Match = 1;
    
-   if (reg.osearch)
+#if SLANG_VERSION < 20000
+   OSearch_Ok = reg.osearch;
+#else
+   (void) SLregexp_get_hints (reg, &flags);
+   OSearch_Ok = (0 != (flags & SLREGEXP_HINT_OSEARCH));
+#endif
+   
+   if (OSearch_Ok)
      {
-	SLsearch_init ((char *) reg.pat, 1, Case_Sensitive, &Search_St);
+#if SLANG_VERSION < 20000
+	SLsearch_init (pattern, 1, Case_Sensitive, &Search_St);
+#else
+	if (NULL == (Search_St = SLsearch_new ((unsigned char *)pattern, Case_Sensitive ? 0 : SLSEARCH_CASELESS)))
+	  exit_error ("SLsearch_new failed");
+#endif
+	Search_St_Key_Length = strlen (pattern);
      }
+#if SLANG_VERSION < 20000
    else if (reg.must_match)
      {
 	SLsearch_init ((char *) reg.must_match_str, 1, Case_Sensitive, &Search_St);
+	Search_St_Key_Length = strlen ((char *) reg.must_match_str);
      }
-   else Must_Match = 0;
-   
-   
+   else
+#endif
+     Must_Match = 0;
+
    if (argc == 0)
      {
 	if (Fixed_Len_Mode) File_Fp = stdin;
