@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -171,6 +171,8 @@ static void unix_update_close (void)
 
 static void handle_interrupt (void)
 {
+   (void) jed_handle_interrupt ();
+
    if (Jed_Sig_Exit_Fun != NULL)
      (*Jed_Sig_Exit_Fun) ();
      
@@ -351,18 +353,22 @@ unsigned char sys_getkey (void) /*{{{*/
 #else
    int all = 0;
 #endif
+   int kb;
 
    while (1)
      {
 	handle_interrupt ();
 
-	if (SLKeyBoard_Quit) break;
+	kb = SLKeyBoard_Quit;
+	if (kb)
+	  break;
 
 	/* sleep for 45 second and try again */
 	if (sys_input_pending(&n, all) > 0)
 	  break;
-
-	if (SLKeyBoard_Quit)
+	
+	kb = SLKeyBoard_Quit;
+	if (kb)
 	  break;
 
 	/* update status line in case user is displaying time */
@@ -373,7 +379,8 @@ unsigned char sys_getkey (void) /*{{{*/
 	  }
      }
 
-   if (SLKeyBoard_Quit)
+   kb = SLKeyBoard_Quit;
+   if (kb)
      {
 	SLKeyBoard_Quit = 0;
 	return Jed_Abort_Char;
@@ -680,7 +687,7 @@ void sys_suspend (void) /*{{{*/
 /*}}}*/
 
 /* returns 0 if file does not exist, 1 if it is not a dir, 2 if it is */
-int sys_chmod(char *file, int what, int *mode, short *uid, short *gid) /*{{{*/
+int sys_chmod(char *file, int what, mode_t *mode, uid_t *uid, gid_t *gid) /*{{{*/
 {
    struct stat buf;
    int m;
@@ -690,12 +697,12 @@ int sys_chmod(char *file, int what, int *mode, short *uid, short *gid) /*{{{*/
 	if (-1 == chmod(file, *mode))
 	  goto return_error;
 
-	if (-1 == chown(file, (uid_t) *uid, (uid_t) *gid))
+	if (-1 == chown(file, *uid, *gid))
 	  {
 	     if (-1 == stat (file, &buf))
 	       goto return_error;
 
-	     (void) chown(file, (uid_t) buf.st_uid, (uid_t) *gid);
+	     (void) chown(file, buf.st_uid, *gid);
 	  }
 	return 0;
      }
@@ -722,7 +729,7 @@ int sys_chmod(char *file, int what, int *mode, short *uid, short *gid) /*{{{*/
 # endif
 #endif
 
-   *mode = m & 0777;
+   *mode = m & 07777;
 
    if (S_ISDIR(m)) return (2);
    return(1);
@@ -767,7 +774,7 @@ int sys_findfirst(char *the_file) /*{{{*/
    char *f, the_path[JED_MAX_PATH_LEN+2], *file;
    unsigned int len;
 
-   file = expand_filename(the_file);
+   file = jed_standardize_filename_static(the_file);
    f = extract_file(file);
    len = strlen (file);
    if (len >= JED_MAX_PATH_LEN)
@@ -822,7 +829,7 @@ static DIR *Dirp;
 
 /* These routines should be fixed to work with wildcards like the DOS and VMS
  * versions do.  Unfortunately, it is a sad fact of life that Unix does NOT
- * support wildcards.  Strangely enough, most Unix users are totally unaware
+ * support wildcards.  Strangely enough, many Unix users are totally unaware
  * of this fact.
  *
  * To add wild card support, I need to use the regular expression package.
@@ -879,7 +886,7 @@ int sys_findfirst(char *the_file) /*{{{*/
 {
    char *f, *file;
 
-   file = expand_filename(the_file);
+   file = jed_standardize_filename_static(the_file);
    safe_strcpy (Found_Dir, file, sizeof (Found_Dir));
    f = extract_file(Found_Dir);
    safe_strcpy (Found_File, f, sizeof (Found_File));
