@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -150,7 +150,8 @@ static int getkey_ast(int not_used) /*{{{*/
      {
 	if (c == (unsigned int) Jed_Abort_Char)
 	  {
-	     if (Ignore_User_Abort == 0) SLang_Error = 2;
+	     if (Ignore_User_Abort == 0) 
+	       SLang_set_error (SL_USER_BREAK);
 	     SLKeyBoard_Quit = 1;
 	  }
 
@@ -415,18 +416,36 @@ unsigned char sys_getkey() /*{{{*/
 int sys_input_pending(int *secs, int unused) /*{{{*/
 {
    unsigned long daytim[2];
-   
+   unsigned long EFstate;
+   float delay = 0.1;
+
    if (Batch) return(0);
    if (Input_Buffer_Len) return(Input_Buffer_Len);
    
    if (X_Input_Pending_Hook != NULL) 
      {
 	if ((*X_Input_Pending_Hook) ()) return 1;
-	/* I need to make this work with DECWIndows */
-	return 0;
+   	if (*secs == 0) return 0;
+   	daytim[1] = 0xFFFFFFFF;
+   	daytim[0] = -(*secs * 1000 * 1000);   /* 1000 * 1000 is a tenth of a sec */
+	sys$setimr(Timer_Event_Flag, daytim, 0, 1);
+   	/* Loop until a key is pressed or the timer event flag gets set */
+	while (1)
+	  {
+	     lib$wait(&delay);	/* so we don't chew up so much cpu time */
+	     if ((*X_Input_Pending_Hook) ())
+	       {
+		  sys$cantim(1, 3);
+		  return 1;
+	       } 
+	     else 
+	       {
+		  if (sys$readef(Timer_Event_Flag, &EFstate) == SS$_WASSET) 
+		    return 0;
+	       }
+	  }
      }
    
-   
    if (*secs) 
      {
 	/* takes a quad word time.  If negative, use a relative time. */
@@ -781,13 +800,16 @@ int vms_parse_file(char *old) /*{{{*/
 
 /*}}}*/
 
-char *expand_filename(char *file) /*{{{*/
+char *jed_standardize_filename_static(char *file) /*{{{*/
 {
    char *p, *p1, *dir;
    static char work [1024];
    int len;
 
-   strcpy(work, get_cwd()); strcat(work, file); file = work;
+   if (NULL != strstr (file, "::"))
+     goto ugly;
+
+   strcpy(work, jed_get_cwd()); strcat(work, file); file = work;
 
     /*  start at end and look for ']' then look for ':' */
     if (0 == (len = strlen(file))) return(file);
@@ -841,6 +863,8 @@ char *expand_filename(char *file) /*{{{*/
 	 p++;
       }
    *p1 = 0;
+
+   ugly:
    
    switch(vms_parse_file(file))
       {
@@ -866,6 +890,16 @@ char *expand_filename(char *file) /*{{{*/
 
 /*}}}*/
 
+char *jed_standardize_filename (char *file)
+{
+   return SLmake_string (jed_standardize_filename_static (file));
+}
+
+char *jed_expand_filename (char *file)
+{
+   return jed_standardize_filename (file);
+}
+
 int vms_expand_filename(char *file,char *expanded_file) /*{{{*/
 {
     static int context = 0;
@@ -953,7 +987,7 @@ int sys_findfirst(char *thefile) /*{{{*/
 	while (*f && (*f != '.')) f++;
 	if (*f) VMS_Star = "*"; else VMS_Star = s2;
 
-	file = expand_filename(thefile);
+	file = jed_standardize_filename_static(thefile);
 	f = f1 = extract_file(file);
 	if (VMS_Star == s2)
 	  {
@@ -982,7 +1016,7 @@ int sys_findfirst(char *thefile) /*{{{*/
 #include <stat.h>
 
 /* returns 0 if file does not exist, 1 if it is not a dir, 2 if it is */
-int sys_chmod(char *file, int what, int *mode, short *uid, short *gid) /*{{{*/
+int sys_chmod(char *file, int what, mode_t *mode, uid_t *uid, gid_t *gid) /*{{{*/
 {
    struct stat buf;
    int m;
@@ -1007,9 +1041,9 @@ int sys_chmod(char *file, int what, int *mode, short *uid, short *gid) /*{{{*/
 /*}}}*/
 
 #if (__VMS_VER < 70000000)
-int rmdir (char *d) /*{{{*/
+int rmdir (const char *d) /*{{{*/
 {
-   return(-1);
+   return -1;
 }
 
 /*}}}*/
