@@ -1,4 +1,4 @@
-/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -36,9 +36,9 @@
 #include "misc.h"
 #include "cmds.h"
 #include "sig.h"
-#include "vterm.h"
 #include "win32.h"
 #include "colors.h"
+/* #include "resource.h" */  /* Where is it? */
 
 #if JED_HAS_SUBPROCESSES
 # include "jprocess.h"
@@ -65,6 +65,10 @@ int NumLock_Is_Gold = 0;
 int PC_Alt_Char = 27;
 int PC_Fn_Char = 0;
 
+/* Last Charset used: keep track of input language changes */
+static DWORD JedCharSet = DEFAULT_CHARSET;
+static DWORD JedACP = CP_ACP;
+
 #define SPACE_CHAR        (32 | (JNORMAL_COLOR << 8))
 
 #define MAX_MENU_ID       256
@@ -456,11 +460,51 @@ static SLang_Intrin_Fun_Type Jed_WinGUI_Table[] =
 #undef S
 #undef I
 
-static void _putkey (char c)
+static int wjed_has_unicode(void)
 {
-   buffer_keystring (&c, 1);
+   static int unicode = -1;
+
+   if (unicode == -1)
+     unicode = SLsmg_is_utf8_mode();
+   return unicode;
 }
 
+static void _putkey (WCHAR wc)
+{
+#if SLANG_VERSION >= 20000
+   if (wjed_has_unicode ())
+     {
+        unsigned char buf[SLUTF8_MAX_MBLEN+1], *b;
+        b = SLutf8_encode((SLwchar_Type)wc, buf, SLUTF8_MAX_MBLEN);
+        buffer_keystring ((char *)buf, b-buf);
+     }
+   else
+#endif
+     {
+        /* 0xE0 is prefix for cursor keys, but also &agrave; on ANSI 
+         * charset. The charmap has 0xE0 0xE0 to &agrave; */
+	unsigned char ch = (unsigned char) wc;
+        if (ch == 0xE0) buffer_keystring ((char *)&ch, 1);
+        buffer_keystring ((char *)&ch, 1);
+     }
+}
+
+static void send_key_sequence (unsigned int n, unsigned char c1, unsigned char c2, 
+			       unsigned char c3, unsigned char c4)
+{
+   unsigned char buf[4];
+
+   if (n > 4)
+     exit_error ("n>4 in send_key_sequence", 0);
+
+   buf[0] = c1;
+   buf[1] = c2;
+   buf[2] = c3;
+   buf[3] = c4;
+   buffer_keystring((char *)buf, n);
+}
+
+
 /* Getting defaults from INI file */
 static COLORREF msw_get_color(char *s, char *dflt)
 {
@@ -621,9 +665,17 @@ static int select_font(char *fontname, int fontheight, int fontbold)
 
    font = CreateFont(height, 0, 0, 0,
                      weight, 0, 0, 0,
-                     DEFAULT_CHARSET, 0, 0, 0, FIXED_PITCH,
+                     JedCharSet, 0, 0, 0, FIXED_PITCH,
                      fontname);
 
+   if (font == NULL)
+     {
+	font == CreateFont(height, 0, 0, 0,
+			   weight, 0, 0, 0,
+			   DEFAULT_CHARSET, 0, 0, 0, FIXED_PITCH,
+			   fontname);
+     }
+
    if (font == NULL)
      {
 	/* This should never happen as we've told CreateFont to use the default */
@@ -748,6 +800,7 @@ static void init_instance (void)
 #if !defined(__WIN32__) || !JED_HAS_SUBPROCESSES
    SetTimer(This_Window.w, 42, 30000, NULL);     /* used for updating display time */
 #endif
+   JedACP = GetACP();
 }
 
 static int msw_init_term (void)
@@ -825,101 +878,201 @@ static void blank_rect(int x1, int y1, int x2, int y2)
 #endif
 }
 
-/* This routine assumes that cursor is in the correct location.  The
- cursor is placed at the end of the string. */
-static void tt_write(char *s, int n)
+static void msw_reverse_video(int color)
 {
-   vterm_write_nchars (s, (unsigned int) n);
+   if ((color < 0) || (color >= JMAX_COLORS))
+     return;
+ 
+   This_Window.current_color = &JColors[color];
+}
+
+static int smg_read_at(int row, int col, SLsmg_Char_Type *s, unsigned int n)
+{
+   int saverow, savecol;
+   unsigned int rc;
+        
+   saverow = SLsmg_get_row();
+   savecol = SLsmg_get_column();
+   SLsmg_gotorc(row, col);
+   rc = SLsmg_read_raw (s, n);
+   SLsmg_gotorc(saverow, savecol);
+   return rc;
+}
+ 
+static void _tt_writeW(WCHAR *s, int n, int color)
+{
+   if (color < 0 || color > JMAX_COLORS)
+     color = 0;
 
    get_dc();
-
-   if (This_Window.cursor_showing) hide_cursor();
    SelectObject(This_Window.hdc, This_Window.font);
-   SetTextColor(This_Window.hdc, This_Window.current_color->fg);
-   SetBkColor(This_Window.hdc, This_Window.current_color->bg);
-   TextOut(This_Window.hdc, This_Window.cursor_col * This_Window.font_width, This_Window.cursor_row * This_Window.font_height, s, n);
-
+   SetTextColor(This_Window.hdc, JColors[color].fg);
+   SetBkColor(This_Window.hdc, JColors[color].bg);
+   TextOutW(This_Window.hdc, This_Window.cursor_col * This_Window.font_width, This_Window.cursor_row * This_Window.font_height, s, n);
    This_Window.cursor_col += n;
 
    if (This_Window.cursor_col >= MSW_Screen_Cols) This_Window.cursor_col = MSW_Screen_Cols - 1;
+   release_dc();
+}
 
+# define LEAD_OFFSET      (0xD800 - (0x10000 >> 10))
+static int decode_utf32 (unsigned int u, WCHAR *buf)
+{
+   if (u < 0xFFFF)
+     {
+	*buf = u;
+	return 1;
+     }
+    else
+      {
+	 *(buf)   = LEAD_OFFSET + (u >> 10);
+	 *(buf+1) = 0xDC00 + (u & 0x3FF);
+	 return 2;
+      }
+}
+
+/* SLsmg_Char_Type decode/compare: the char internals changed for SLang2 */
+#if SLANG_VERSION >= 20000
+   
+static int decode_smgchar(SLsmg_Char_Type *s, WCHAR *buf)
+{
+   if (s->nchars > 0) 
+     return decode_utf32(s->wchars[0], buf);
+
+   *buf = (WCHAR)' ';
+   return 1;
+}
+
+# define SLSMGCHAR_EQUAL(o, n) \
+   ( ((o)->nchars == (n)->nchars) \
+	&& ((o)->color  == (n)->color) \
+	&& (0 == memcmp((o)->wchars, (n)->wchars, (n)->nchars * sizeof(SLwchar_Type))))
+# define SLSMGCHAR_SET_CHAR(sc, _char) do { (sc).nchars = 1; (sc).wchars[0] = (_char); } while (0)
+# define SLSMGCHAR_SET_COLOR(sc, _color) (sc).color = (_color)
+#else
+
+static int decode_smgchar(SLsmg_Char_Type *s, WCHAR *buf)
+{
+   *buf = SLSMG_EXTRACT_CHAR(*s);
+   return 1;
+}
+
+# define SLSMGCHAR_EQUAL(o, n) ((o) == (n))
+# define SLSMGCHAR_SET_CHAR(sc, _char) (sc) = (sc) & 0xFF00 + (_char)
+# define SLSMGCHAR_SET_COLOR(sc, _color) (sc) = ((sc) & 0xFF) + ((_color)<<8)
+
+/* define this to allow using the same code for slang 1 */
+# define SLSMG_COLOR_MASK 0xFF
+
+#endif				       /* SLANG_VERSION >= 20000 */
+
+static void msw_write_smgchars(SLsmg_Char_Type *s, SLsmg_Char_Type *smax)
+{
+   WCHAR buf[512];
+   WCHAR *b, *bmax;
+   int oldcolor, color;
+   
+   b = buf;
+   bmax = b + sizeof (buf)/sizeof(buf[0]);
+
+   hide_cursor();
+
+   oldcolor = (SLSMG_EXTRACT_COLOR(*s) & SLSMG_COLOR_MASK);
+   color = 0;
+   get_dc();
+   for (; s < smax; s++)
+     {
+        color = (SLSMG_EXTRACT_COLOR(*s) & SLSMG_COLOR_MASK);
+        if (oldcolor != color ||
+            b >= bmax-1) /* always leave room for 1 more 1 char (for surrogates) */
+          {
+             _tt_writeW(buf, b - buf, oldcolor);
+             b = buf;
+             msw_reverse_video(color);
+             oldcolor = color;
+          }
+        b += decode_smgchar(s, b);
+   }
+   if (b != buf)
+     _tt_writeW(buf, b - buf, color);
+
+   release_dc();
+}
+
+
+static void msw_write_smgchar(SLsmg_Char_Type *s)
+{
+   WCHAR buf[2];
+   int l;
+
+   get_dc();
+   l = decode_smgchar(s, buf);
+   _tt_writeW (buf, l, SLSMG_EXTRACT_COLOR(*s) & SLSMG_COLOR_MASK);
    release_dc();
 }
 
 static void hide_cursor(void)
 {
-   SLsmg_Char_Type *s;
-   char ch;
+   SLsmg_Char_Type sc;
    int col = This_Window.vis_curs_col, row = This_Window.vis_curs_row;
-   Color_Type *color;
+   int savecur;
 
-   if ((This_Window.cursor_showing == 0)
-       || (VTerm_Display == NULL))
+   if (This_Window.cursor_showing == 0)
      return;
-
    This_Window.cursor_showing = 0;
 
-   s = VTerm_Display[row] + col; 
-   ch = *s & 0xFF;
-
-   get_dc();
-   color = &JColors[*s >> 8];
-   SelectObject(This_Window.hdc, This_Window.font);
-   SetTextColor(This_Window.hdc, color->fg);
-   SetBkColor(This_Window.hdc, color->bg);
-   TextOut(This_Window.hdc, This_Window.cursor_col * This_Window.font_width, This_Window.cursor_row * This_Window.font_height, &ch, 1);
-   release_dc();
+   if (smg_read_at(row, col, &sc, 1) == 0)
+     {
+	SLSMGCHAR_SET_CHAR(sc, ' ');
+	SLSMGCHAR_SET_COLOR(sc, JNORMAL_COLOR);
+     }
+   savecur = This_Window.cursor_col;
+   msw_write_smgchar(&sc);
+   This_Window.cursor_col = savecur;
 }
 
 
 static void show_cursor(void)
 {
-   SLsmg_Char_Type *s;
-   char ch;
-   int c, r;
-   Color_Type *curs_color;
-   RECT rc;
-
-   if (VTerm_Display == NULL)
-     return;
+   SLsmg_Char_Type sc;
+   int color;
 
    if (This_Window.cursor_showing) hide_cursor();
 
    This_Window.cursor_showing = 1;
+   This_Window.vis_curs_row = This_Window.cursor_row;
+   This_Window.vis_curs_col = This_Window.cursor_col;
 
-   r = This_Window.vis_curs_row = This_Window.cursor_row;
-   c = This_Window.vis_curs_col = This_Window.cursor_col;
-
-   s = VTerm_Display[r] + c;
-   ch = *s & 0xFF;
-
-   ch = (char) (*s & 0xFF);
    if (CBuf && CBuf->flags & OVERWRITE_MODE)
-     curs_color = &JColors[JCURSOROVR_COLOR];
+     color = JCURSOROVR_COLOR;
    else
-     curs_color = &JColors[JCURSOR_COLOR];
+     color = JCURSOR_COLOR;
 
-   get_dc();
    if (This_Window.focus)
      {
-	SelectObject(This_Window.hdc, This_Window.font);
-	SetTextColor(This_Window.hdc, curs_color->fg);
-	SetBkColor(This_Window.hdc, curs_color->bg);
-	TextOut(This_Window.hdc, This_Window.cursor_col * This_Window.font_width, This_Window.cursor_row * This_Window.font_height, &ch, 1);
+	int savecur;
+	if (smg_read_at(This_Window.cursor_row, This_Window.cursor_col, &sc, 1) == 0)
+	  SLSMGCHAR_SET_CHAR(sc, ' ');
+	SLSMGCHAR_SET_COLOR(sc, color);
+        savecur = This_Window.cursor_col;
+        msw_write_smgchar(&sc);
+        This_Window.cursor_col = savecur;
      }
    else
      {
+	RECT rc;
 	rc.left = This_Window.cursor_col * This_Window.font_width;
 	rc.top = This_Window.cursor_row * This_Window.font_height;
 	rc.right = rc.left + This_Window.font_width;
 	rc.bottom = rc.top + This_Window.font_height;
-	FrameRect(This_Window.hdc, &rc, curs_color->hbrBG);
+	get_dc();
+	FrameRect(This_Window.hdc, &rc, JColors[color].hbrBG);
+	release_dc();
      }
-
-   release_dc();
 }
 
 
+
 void sys_suspend(void)
 {
    ShowWindow(This_Window.w, SW_MINIMIZE);
@@ -1000,7 +1153,6 @@ static void msw_goto_rc(int r, int c)
    This_Window.cursor_row = r;
    This_Window.cursor_col = c;
 
-   vterm_goto_rc (r, c);
    if (!Performing_Update) show_cursor();
    release_dc();
 }
@@ -1008,7 +1160,6 @@ static void msw_goto_rc(int r, int c)
 static void msw_del_eol(void)
 {
    blank_rect(This_Window.cursor_col, This_Window.cursor_row, MSW_Screen_Cols, This_Window.cursor_row + 1);
-   vterm_del_eol ();
 }
 
 static void msw_delete_nlines(int n)
@@ -1020,7 +1171,6 @@ static void msw_delete_nlines(int n)
    if (r1 <= r2 - n) copy_rect(0, r1 + n, MSW_Screen_Cols, r2 + 1, 0, r1);
    r2++;
    blank_rect(0, r2 - n, MSW_Screen_Cols, r2);
-   vterm_delete_nlines (n);
 }
 
 #if 0				       /* UNUSED */
@@ -1066,17 +1216,6 @@ static void msw_putchar(char ch)
 
 #endif				       /* UNUSED */
 
-static int Rev_Vid_Flag;
-static void msw_reverse_video(int color)
-{
-   if ((color < 0) || (color > JMAX_COLORS))
-     return;
-
-   Rev_Vid_Flag = color;
-   This_Window.current_color = &JColors[color];
-   vterm_reverse_video (color);
-}
-
 static void msw_normal_video(void)
 {
    msw_reverse_video (JNORMAL_COLOR);
@@ -1094,7 +1233,6 @@ static void msw_cls(void)
     * from the desired background.  So patch it up with blank_rect.
     */
    blank_rect (0, 0, MSW_Screen_Cols, MSW_Screen_Rows);
-   vterm_cls ();
 }
 
 static void msw_beep(void)
@@ -1127,76 +1265,35 @@ static void msw_reverse_index(int n)
    if (r2 >= r1 + n) copy_rect(0, r1, MSW_Screen_Cols, r2 - n + 1, 0, r1 + n);
 
    blank_rect(0, r1, MSW_Screen_Cols, r1 + n);
-   vterm_reverse_index (n);
 }
 
 static void msw_write_string (char *s)
 {
    get_dc();
-   tt_write(s, strlen(s));
+   /* tt_write(s, strlen(s)); */
    if (!Performing_Update) show_cursor ();
    release_dc();
 }
 
-static void send_attr_str (SLsmg_Char_Type *s, SLsmg_Char_Type *smax)
-{
-   unsigned char out[256], ch, attr, *p, *pmax;
-   register SLsmg_Char_Type sh;
-   
-   p = out;
-   pmax = out + (sizeof(out)-1);
-   get_dc();
-   while (s < smax)
-     {
-	sh = *s++;
-	ch = sh & 0xFF;
-	attr = sh >> 8;
-	if ((attr == 0) && (Rev_Vid_Flag != 0))
-	  {
-	     if (p != out)
-	       {
-		  *p = 0;
-		  msw_write_string ((char *) out);
-		  p = out;
-	       }
-	     msw_normal_video();
-	     /* Rev_Vid_Flag = 0; */
-	  }
-	else if ((attr != 0) && (Rev_Vid_Flag != attr))
-	  {
-	     if (p != out)
-	       {
-		  *p = 0;
-		  msw_write_string ((char *) out);
-		  p = out;
-	       }
-	     msw_reverse_video(attr);
-	     /* Rev_Vid_Flag = 1; */
-	  }
-	*p = ch;
-	if (p < pmax)
-	  p++;
-     }
-   *p = 0;
-   if (p != out) msw_write_string ((char *) out);
-   /* if (Rev_Vid_Flag) tt_normal_video(); */
-   release_dc();
-}
-
 static void msw_smart_puts(SLsmg_Char_Type *neww, SLsmg_Char_Type *oldd, int len, int row)
 {
-   if (0 == memcmp ((char *) neww, (char *) oldd, len * sizeof (SLsmg_Char_Type)))
-     return;
+   int col;
 
-   msw_goto_rc (row, 0);
-   send_attr_str (neww, neww + len);
+   col = 0;
+   while ((col < len) && SLSMGCHAR_EQUAL(&neww[col], &oldd[col]))
+     col++;
+
+   if (col < len) 
+     {
+	msw_goto_rc (row, col);
+	msw_write_smgchars(&neww[col], &neww[len]);
+     }
 }
 
 static void msw_set_scroll_region(int r1, int r2)
 {
    This_Window.scroll_r1 = r1;
    This_Window.scroll_r2 = r2;
-   vterm_set_scroll_region (r1, r2);
 }
 
 static void msw_reset_scroll_region (void)
@@ -1208,13 +1305,10 @@ static void msw_reset_scroll_region (void)
 
 static int msw_reset_video (void)
 {
-   if (VTerm_Display != NULL)
-     {
-	msw_reset_scroll_region ();
-	msw_goto_rc (0, 0);
-	msw_normal_video ();
-     }
-   return vterm_reset_display ();
+   msw_reset_scroll_region ();
+   msw_goto_rc (0, 0);
+   msw_normal_video ();
+   return 0;
 }
 
 static int msw_init_video (void)
@@ -1225,8 +1319,7 @@ static int msw_init_video (void)
 	MSW_Screen_Cols = 80;
 	MSW_Screen_Rows = 24;
      }
-
-   return vterm_init_display (MSW_Screen_Rows, MSW_Screen_Cols);
+   return 0;
 }
 
 static int msw_flush_output (void)
@@ -1234,52 +1327,6 @@ static int msw_flush_output (void)
    return 0;
 }
 
-
-static void msw_get_terminfo (void)
-{
-   SLsmg_Term_Type tt;
-
-   MSW_Screen_Cols = 80;
-   MSW_Screen_Rows = 24;
-
-   (void) jed_add_init_slang_hook (msw_init_slang);
-
-   /* init hooks */
-   X_Update_Open_Hook = msw_update_open;
-   X_Update_Close_Hook = msw_update_close;
-   X_Suspend_Hook = msw_suspend;
-   /*   X_Argc_Argv_Hook = X_eval_command_line; */
-   X_Define_Keys_Hook = msw_define_xkeys;
-   SLang_Interrupt = msw_check_kbd;
-   X_Init_Term_Hook = msw_init_term;
-
-   /* Set this so that main will not try to read from stdin.  It is quite
-    * likely that this is started from a menu or something.
-    */
-   Stdin_Is_TTY = -1;
-   
-   memset ((char *) &tt, 0, sizeof (SLsmg_Term_Type));
-
-   tt.tt_normal_video = msw_normal_video;
-   tt.tt_set_scroll_region = msw_set_scroll_region;
-   tt.tt_goto_rc = msw_goto_rc;
-   tt.tt_reverse_index = msw_reverse_index;
-   tt.tt_reset_scroll_region = msw_reset_scroll_region;
-   tt.tt_delete_nlines = msw_delete_nlines;
-   tt.tt_cls = msw_cls;
-   tt.tt_del_eol = msw_del_eol;
-   tt.tt_smart_puts = msw_smart_puts;
-   tt.tt_flush_output = msw_flush_output;
-   tt.tt_reset_video = msw_reset_video;
-   tt.tt_init_video = msw_init_video;
-   
-   tt.tt_screen_rows = &MSW_Screen_Rows;
-   tt.tt_screen_cols = &MSW_Screen_Cols;
-   tt.tt_term_cannot_scroll = &MSW_Term_Cannot_Scroll;
-
-   SLsmg_set_terminal_info (&tt);
-}
-
 static char *convert_color (char *c, char *buf)
 {
    unsigned long ul;
@@ -1299,7 +1346,7 @@ static char *convert_color (char *c, char *buf)
 }
    
 
-static void msw_set_color (int i, char *what, char *fg, char *bg)
+static JX_SETXXX_RETURN_TYPE msw_set_color (int i, char *what, char *fg, char *bg)
 {
    int r, g, b;
    char buf[64];
@@ -1308,7 +1355,7 @@ static void msw_set_color (int i, char *what, char *fg, char *bg)
    HBRUSH bgbr;
 
    if ((i < 0) || (i > JMAX_COLORS))
-     return;
+     return JX_SETXXX_RETURN_VAL;
 
    (void) what;
 
@@ -1355,19 +1402,18 @@ static void msw_set_color (int i, char *what, char *fg, char *bg)
 #if SLANG_VERSION >= 10306
    SLsmg_touch_screen ();
 #endif
+
+   return JX_SETXXX_RETURN_VAL;
+
    /* InvalidateRect(This_Window.w, NULL, FALSE); */
 }
 
 
 static void cover_exposed_area (int x, int y, int width, int height)
 {
-   SLsmg_Char_Type *s, *smax;
+   SLsmg_Char_Type *s;
    int row, save_row, save_col, max_col, max_row, col;
-
-   /* VTerm_Display can be NULL here, handle this
-    * cpbotha@bigfoot.com
-    */
-   if (VTerm_Display==NULL) return;
+   int width_chars, len;
 
    Performing_Update++;
    hide_cursor ();
@@ -1376,20 +1422,26 @@ static void cover_exposed_area (int x, int y, int width, int height)
    col = x / This_Window.font_width;
    row = y / This_Window.font_height;
 
-   max_col = 2 + col + width / This_Window.font_width;
+   width_chars = 2 + width / This_Window.font_width;
+   max_col = col + width_chars;
    max_row = 2 + row + height / This_Window.font_height;
    if (max_col > MSW_Screen_Cols) max_col = MSW_Screen_Cols;
    if (max_row > MSW_Screen_Rows) max_row = MSW_Screen_Rows;
 
+   if (NULL == (s = (SLsmg_Char_Type *)SLmalloc(width_chars*sizeof(*s))))
+     goto done;
+
    while (row < max_row)
      {
 	msw_goto_rc (row, col);
-	s = VTerm_Display [row];
-	smax = s + max_col;
-	send_attr_str(s + col, smax);
+        len = smg_read_at(row, col, s, width_chars);
+        msw_write_smgchars(s, s + len);
 	row++;
      }
+   SLfree((char *)s);
    msw_goto_rc (save_row, save_col);
+   
+   done:
    Performing_Update--;
 
    show_cursor ();
@@ -1475,8 +1527,7 @@ static void push_mouse_event(int button, int x, int y, int state, int type)
 	return;
      }
 
-   buffer_keystring ("\033\000", 2);
-   _putkey(id);
+   send_key_sequence(3, 033, 0, id, 0);
 }
 
 static void push_wm_mouse_event (int button, int x, int y, int state, int type)
@@ -1522,6 +1573,10 @@ static void msw_select_font (char *fontname, int *height, int *bold)
      {
 	jed_init_display ();
 	jed_redraw_screen (1);
+	/* update font data, As I need it if the user changes Language */
+	strcpy(Font_Name, fontname);
+	Font_Height = *height;
+	Font_Bold = *bold;	
 	return;
      }
    jed_verror ("Unable to allocate font %s", fontname);
@@ -1561,7 +1616,7 @@ static void msw_cleanup(void)
      if (JColors[i].hbrBG) DeleteObject(JColors[i].hbrBG);
 }
 
-#if !defined(__WIN32__) || !defined(JED_HAS_SUBPROCESSES)
+#if !defined(__WIN32__) || !JED_HAS_SUBPROCESSES
 int sys_input_pending(int *tsecs, int unused)
 {
    DWORD t = GetTickCount() + *tsecs * 100L;
@@ -1672,7 +1727,7 @@ unsigned char sys_getkey(void)
 
 static int Ignore_Wchar_Message;
 
-static char f_keys[4][12] =
+static unsigned char f_keys[4][12] =
 {
      { 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 133, 134 },
      { 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 135, 136 },
@@ -1680,7 +1735,7 @@ static char f_keys[4][12] =
      { 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 139, 140 }
 };
 
-static char small_keypad_keys[4][13] =
+static unsigned char small_keypad_keys[4][13] =
 {
      { 'G', 'H', 'I', 0, 'K', 0, 'M', 0, 'O', 'P', 'Q', 'R', 'S' },   /* normal */
      { '0', '1', '2', 0, '3', 0, '4', 0, '5', '6', '7', '8', '9' },   /* shift */
@@ -1688,7 +1743,7 @@ static char small_keypad_keys[4][13] =
      { 151, 152, 153, 0, 155, 0, 157, 0, 159, 160, 161, 162, 163 }   /* alt */
 };
 
-static char num_keypad_keys[4][13] =
+static unsigned char num_keypad_keys[4][13] =
 {
      { 'w', 'x', 'y', 0, 't', 'u', 'v', 0, 'q', 'r', 's', 'p', 'n' },
      { '0', '1', '2', 0, '3',  0 , '4', 0, '5', '6', '7', '8', '9' },
@@ -1700,7 +1755,7 @@ static LRESULT CALLBACK process_key_down (HWND hWnd, UINT msg, WPARAM wParam, LP
 {
    unsigned int key_state = 0;
    unsigned int scan;
-   char prefix, c1;
+   unsigned char prefix, c1;
    int i, state;
 
    if (GetKeyState(VK_CONTROL) & 0x8000) key_state |= KEY_CONTROL;
@@ -1723,19 +1778,11 @@ static LRESULT CALLBACK process_key_down (HWND hWnd, UINT msg, WPARAM wParam, LP
 	if ((key_state & KEY_ALT) == 0)
 	  {
 	     if ((key_state & KEY_CONTROL) != 0)
-	       {
-		  _putkey(PC_Alt_Char);
-		  _putkey('@');
-	       }
+	       send_key_sequence(2, PC_Alt_Char, '@', 0, 0);
 	     else if ((key_state & KEY_SHIFT) != 0)
-	       {
-		  _putkey(PC_Alt_Char);
-		  _putkey('?');
-	       }
+	       send_key_sequence(2, PC_Alt_Char, '?', 0, 0);
 	     else
-	       {
-		  _putkey (127);
-	       }
+	       send_key_sequence(1, 127, 0, 0, 0);
 	  }
 	Ignore_Wchar_Message = 1;
 	return 0;
@@ -1763,7 +1810,7 @@ static LRESULT CALLBACK process_key_down (HWND hWnd, UINT msg, WPARAM wParam, LP
 	    && (key_state & KEY_CONTROL))
 	  {
 	     Ignore_Wchar_Message = 1;
-	     _putkey (31);	       /* ^_ */
+	     send_key_sequence(1, 31, 0, 0, 0); /* ^_ */
 	     return 0;
 	  }
 	return DefWindowProc(hWnd, msg, wParam, lParam);
@@ -1827,8 +1874,7 @@ static LRESULT CALLBACK process_key_down (HWND hWnd, UINT msg, WPARAM wParam, LP
  	if (key_state & (KEY_CONTROL | KEY_ALT))
  	  {
 	     Ignore_Wchar_Message = 1;
- 	     _putkey (0);
- 	     _putkey (num_keypad_keys[state][scan - 0x47]);
+	     send_key_sequence(2, 0, num_keypad_keys[state][scan - 0x47], 0, 0);
  	     return 0;
  	  }
  	else
@@ -1838,8 +1884,7 @@ static LRESULT CALLBACK process_key_down (HWND hWnd, UINT msg, WPARAM wParam, LP
       case 0x11C:            /* KEYPAD ENTER */
 	if (key_state & KEY_ALT)
 	  {
- 	     _putkey(0);
- 	     _putkey(166);
+	     send_key_sequence(2, 0, 166, 0, 0);
 	     Ignore_Wchar_Message = 1;
 	     return 0;
  	  }
@@ -1869,8 +1914,7 @@ static LRESULT CALLBACK process_key_down (HWND hWnd, UINT msg, WPARAM wParam, LP
 	     state = 3;
 	  }
 
-	_putkey (prefix);
-	_putkey (small_keypad_keys[state][scan - 0x147]);
+	send_key_sequence(2, prefix, small_keypad_keys[state][scan - 0x147], 0, 0);
 	Ignore_Wchar_Message = 1;
 	return 0;
 
@@ -1900,22 +1944,20 @@ static LRESULT CALLBACK process_key_down (HWND hWnd, UINT msg, WPARAM wParam, LP
 	       state = 3;
 	  }
 
-	_putkey(PC_Fn_Char);
-	_putkey(f_keys[state][i]);
+	send_key_sequence(2, PC_Fn_Char, f_keys[state][i], 0, 0);
 	Ignore_Wchar_Message = 1;
 	return 0;
      }
 
-   buffer_keystring ("\033O", 2);
-   _putkey (c1);
+   send_key_sequence(3, 033, 'O', c1, 0);
    Ignore_Wchar_Message = 1;
    return 0;
 }
 
 static int load_dropped_files(HDROP hDrop, HWND hwnd)
 {
-   CHAR szFilename[JED_MAX_PATH_LEN];
-   INT iNumFiles;
+   char szFilename[JED_MAX_PATH_LEN];
+   int iNumFiles;
 
    (void) hwnd;
 
@@ -1941,6 +1983,17 @@ LRESULT CALLBACK JEDWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
 
    switch (msg)
      {
+      case WM_INPUTLANGCHANGE:
+	  {
+	     CHARSETINFO csi;
+
+	     JedCharSet = wParam;
+	     TranslateCharsetInfo((DWORD *)wParam, &csi, TCI_SRCCHARSET);
+	     JedACP = csi.ciACP;
+	     (void) msw_select_font (Font_Name, &Font_Height, &Font_Bold);
+	  }
+	return 0;
+
       case WM_CREATE:
 	This_Window.w = hWnd;
 	DragAcceptFiles (hWnd, TRUE);
@@ -2016,7 +2069,7 @@ LRESULT CALLBACK JEDWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
 	  {
 	     if ((int) wParam == (int) Jed_Abort_Char)
 	       {
-		  if (Ignore_User_Abort == 0) SLang_Error = USER_BREAK;
+		  if (Ignore_User_Abort == 0) SLang_set_error(USER_BREAK);
 		  SLKeyBoard_Quit = 1;
 	       }
 	     /* PS 28Apr98: handle shiftTAB same as DOS */
@@ -2024,14 +2077,41 @@ LRESULT CALLBACK JEDWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
 	       {
 		  if (GetKeyState(VK_SHIFT)<0)
 		    {
-		       _putkey(0);
-		       _putkey('o' & 31);
+		       send_key_sequence(2, 0, 'o' & 31, 0, 0);
 		       Ignore_Wchar_Message = 1;
 		       break;
 		    }
 	       }
-	     if (wParam == 0xE0) _putkey(wParam);
-	     _putkey(wParam);
+	     if (!wjed_has_unicode())
+	       _putkey(wParam);
+	     else
+	       {
+		  static char dbcsbuf[3] = "";
+		  WCHAR buf[10];
+		  int i, buflen = 10;
+		  int dbcsl;
+		  
+		  // Here I get ANSI characters, but _putkey() wants UTF-16.
+		  if (!dbcsbuf[0] && IsDBCSLeadByte((unsigned char)wParam)) {
+		     dbcsbuf[0] = (char)wParam;
+		     break;
+		  }
+		  dbcsl = 1;
+		  if (dbcsbuf[0]) 
+		    {
+		       dbcsbuf[1] = (unsigned char) wParam;
+		       dbcsl = 2;
+		    } 
+		  else 
+		    {
+		       dbcsbuf[0] = (unsigned char) wParam;
+		    }
+		  dbcsbuf[2] = 0;
+		  buflen = MultiByteToWideChar(JedACP, 0, dbcsbuf, dbcsl, buf, buflen);
+		  for (i = 0; i < buflen; i++)
+		    _putkey(buf[i]);
+		  dbcsbuf[0] = 0;
+	       }
 	     break;
 	  }
 	return DefWindowProc(hWnd, msg, wParam, lParam);
@@ -2114,7 +2194,7 @@ LRESULT CALLBACK JEDWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
 		  if (is_internal(Menu_Callbacks[wParam]))
 		    {
 		       char buf[512];
-		       _SLsnprintf(buf, sizeof (buf), ". \"%s\" call", Menu_Callbacks[wParam]);
+		       SLsnprintf(buf, sizeof (buf), ". \"%s\" call", Menu_Callbacks[wParam]);
 		       SLang_load_string (buf);
 		    }
 		  else
@@ -2162,6 +2242,22 @@ LRESULT CALLBACK JEDWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
    return 0;
 }
 
+static int pop_hmenu (HMENU *h)
+{
+   int i;
+   
+   if (-1 == SLang_pop_integer (&i))
+     return -1;
+   
+   *h = (HMENU) i;
+   return 0;
+}
+
+static int push_hmenu (HMENU h)
+{
+   return SLang_push_integer ((int) h);
+}
+
 void get_menubar()
 {
    HMENU hmenu = GetMenu(This_Window.w);
@@ -2172,8 +2268,7 @@ void get_menubar()
 	SetMenu(This_Window.w, hmenu);
 	DrawMenuBar(This_Window.w);
      }
-
-   SLang_push_integer((int) hmenu);
+   (void) push_hmenu (hmenu);
 }
 
 void destroy_menubar()
@@ -2194,8 +2289,8 @@ void destroy_menubar()
 void create_popup_menu()
 {
    HMENU hmenu = CreatePopupMenu();
-
-   SLang_push_integer((int) hmenu);
+   
+   (void) push_hmenu (hmenu);
 }
 
 void destroy_menu()
@@ -2204,7 +2299,7 @@ void destroy_menu()
    int count, i;
    int id;
 
-   if (!SLang_pop_integer((int *)&hmenu))
+   if (0 == pop_hmenu (&hmenu))
      {
 	count = GetMenuItemCount(hmenu);
 	for(i = 0; i < count; i++)
@@ -2231,7 +2326,7 @@ void append_menu_item()
    if (!SLpop_string(&callback) &&
        !SLang_pop_integer(&id) &&
        !SLpop_string(&name) &&
-       !SLang_pop_integer((int *)&hmenu))
+       !pop_hmenu (&hmenu))
      {
 	if ((id < 0) || (id >= MAX_MENU_ID))
 	  msg_error("Id is out of range.");
@@ -2264,9 +2359,9 @@ void append_popup_menu()
    char *name = NULL;
    HMENU popup;
 
-   if (!SLang_pop_integer((int *)&popup) &&
+   if (!pop_hmenu(&popup) &&
        !SLpop_string (&name) &&
-       !SLang_pop_integer((int *)&hmenu))
+       !pop_hmenu(&hmenu))
      {
 	if (!AppendMenu(hmenu, MF_STRING | MF_POPUP, (UINT)popup, name))
 	  msg_error("Cannot append popup menu");
@@ -2279,7 +2374,7 @@ void append_separator()
 {
    HMENU hmenu;
 
-   if (!SLang_pop_integer((int *)&hmenu))
+   if (!pop_hmenu (&hmenu))
      {
 	if (!AppendMenu(hmenu, MF_STRING | MF_SEPARATOR, 0, 0))
 	  msg_error("Cannot append separator");
@@ -2298,7 +2393,7 @@ void insert_menu_item()
        !SLang_pop_integer(&idNew) &&
        !SLpop_string(&name) &&
        !SLang_pop_integer(&id) &&
-       !SLang_pop_integer((int *)&hmenu))
+       !pop_hmenu(&hmenu))
      {
 	if ((idNew < 0) || (idNew >= MAX_MENU_ID))
 	  msg_error("Id is out of range");
@@ -2331,10 +2426,10 @@ void insert_popup_menu()
    char *name = NULL;
    HMENU popup;
 
-   if (!SLang_pop_integer((int *)&popup) &&
+   if (!pop_hmenu(&popup) &&
        !SLpop_string (&name) &&
        !SLang_pop_integer(&id) &&
-       !SLang_pop_integer((int *)&hmenu))
+       !pop_hmenu(&hmenu))
      {
 	if (!InsertMenu(hmenu, id, MF_STRING | MF_POPUP | MF_BYCOMMAND, (UINT)popup, name))
 	  msg_error("Cannot insert popup menu");
@@ -2349,7 +2444,7 @@ void insert_separator()
    int   id;
 
    if (!SLang_pop_integer(&id) &&
-       !SLang_pop_integer((int *)&hmenu))
+       !pop_hmenu (&hmenu))
      {
 	if (!InsertMenu(hmenu, id, MF_STRING | MF_SEPARATOR | MF_BYCOMMAND, 0, 0))
 	  msg_error("Cannot insert separator");
@@ -2368,7 +2463,7 @@ void insert_menu_item_pos()
        !SLang_pop_integer(&idNew) &&
        !SLpop_string(&name) &&
        !SLang_pop_integer(&pos) &&
-       !SLang_pop_integer((int *)&hmenu))
+       !pop_hmenu(&hmenu))
      {
 	if ((idNew < 0) || (idNew >= MAX_MENU_ID))
 	  msg_error("Id is out of range.");
@@ -2401,10 +2496,10 @@ void insert_popup_menu_pos()
    char *name = NULL;
    HMENU popup;
 
-   if (!SLang_pop_integer((int *)&popup) &&
+   if (!pop_hmenu(&popup) &&
        !SLpop_string(&name) &&
        !SLang_pop_integer(&pos) &&
-       !SLang_pop_integer((int *)&hmenu))
+       !pop_hmenu(&hmenu))
      {
 	if (!InsertMenu(hmenu, pos, MF_STRING | MF_POPUP | MF_BYPOSITION, (UINT)popup, name))
 	  msg_error("Cannot insert popup menu");
@@ -2419,7 +2514,7 @@ void insert_separator_pos()
    int   pos;
 
    if (!SLang_pop_integer(&pos) &&
-       !SLang_pop_integer((int *)&hmenu))
+       !pop_hmenu(&hmenu))
      {
 	if (!InsertMenu(hmenu, pos, MF_STRING | MF_SEPARATOR | MF_BYPOSITION, 0, 0))
 	  msg_error("Cannot insert separator");
@@ -2432,7 +2527,7 @@ void delete_menu_item()
    int   id;
 
    if (!SLang_pop_integer(&id) &&
-       !SLang_pop_integer((int *)&hmenu))
+       !pop_hmenu(&hmenu))
      {
 	if ((id < 0) || (id >= MAX_MENU_ID))
 	  {
@@ -2460,7 +2555,7 @@ void delete_menu_item_pos()
    int   id;
 
    if (!SLang_pop_integer(&pos) &&
-       !SLang_pop_integer((int *)&hmenu))
+       !pop_hmenu(&hmenu))
      {
 	id = GetMenuItemID(hmenu, pos);
 	if ((id > 0) && (Menu_Callbacks[id] != NULL))
@@ -2472,7 +2567,7 @@ void delete_menu_item_pos()
 	  if (id == -1)
 	  {
 	     popup = GetSubMenu(hmenu, pos);
-	     SLang_push_integer((int)popup);
+	     (void) push_hmenu (popup);
 	     destroy_menu();
 	  }
 
@@ -2489,7 +2584,7 @@ void get_menu_state()
    int   state;
 
    if (!SLang_pop_integer(&id) &&
-       !SLang_pop_integer((int *)&hmenu))
+       !pop_hmenu(&hmenu))
      {
 	if (-1 == (int)(mstate = GetMenuState(hmenu, id, MF_BYCOMMAND)))
 	  msg_error("Cannot get menu state");
@@ -2499,7 +2594,7 @@ void get_menu_state()
 	     if (mstate & MF_ENABLED) state = 1;
 	     if (mstate & MF_CHECKED) state |= 2;
 
-	     SLang_push_integer(state);
+	     (void)SLang_push_integer(state);
 	  }
      }
 }
@@ -2512,7 +2607,7 @@ void get_menu_state_pos()
    int   state;
 
    if (!SLang_pop_integer(&pos) &&
-       !SLang_pop_integer((int *)&hmenu))
+       !pop_hmenu(&hmenu))
      {
 	if (-1 == (int)(mstate = GetMenuState(hmenu, pos, MF_BYPOSITION)))
 	  msg_error("Cannot get menu state");
@@ -2522,7 +2617,7 @@ void get_menu_state_pos()
 	     if (mstate & MF_GRAYED) state = 1;
 	     if (mstate & MF_CHECKED) state |= 2;
 
-	     SLang_push_integer(state);
+	     (void)SLang_push_integer(state);
 	  }
      }
 }
@@ -2534,10 +2629,10 @@ void get_popup_menu()
    HMENU popup;
 
    if (!SLang_pop_integer(&pos) &&
-       !SLang_pop_integer((int *)&hmenu))
+       !pop_hmenu(&hmenu))
      {
 	popup = GetSubMenu(hmenu, pos);
-	SLang_push_integer((int) popup);
+	(void) push_hmenu (popup);
      }
 }
 
@@ -2549,7 +2644,7 @@ void check_menu_item()
 
    if (!SLang_pop_integer(&flag) &&
        !SLang_pop_integer(&id) &&
-       !SLang_pop_integer((int *)&hmenu))
+       !pop_hmenu(&hmenu))
      {
 	if (-1 == (int) CheckMenuItem(hmenu, id, MF_BYCOMMAND | (flag)?MF_CHECKED:MF_UNCHECKED))
 	  msg_error("Menu item does not exist");
@@ -2564,7 +2659,7 @@ void check_menu_item_pos()
 
    if (!SLang_pop_integer(&flag) &&
        !SLang_pop_integer(&pos) &&
-       !SLang_pop_integer((int *)&hmenu))
+       !pop_hmenu(&hmenu))
      {
 	if (!CheckMenuItem(hmenu, pos, MF_BYPOSITION | (flag)?MF_CHECKED:MF_UNCHECKED))
 	  msg_error("Menu item does not exist");
@@ -2579,7 +2674,7 @@ void enable_menu_item()
 
    if (!SLang_pop_integer(&flag) &&
        !SLang_pop_integer(&id) &&
-       !SLang_pop_integer((int *)&hmenu))
+       !pop_hmenu(&hmenu))
      {
 	if (-1 == EnableMenuItem(hmenu, id, MF_BYCOMMAND | (flag)?MF_ENABLED:MF_GRAYED))
 	  msg_error("Menu item does not exist");
@@ -2594,7 +2689,7 @@ void enable_menu_item_pos()
 
    if (!SLang_pop_integer(&flag) &&
        !SLang_pop_integer(&pos) &&
-       !SLang_pop_integer((int *)&hmenu))
+       !pop_hmenu(&hmenu))
      {
 	if (-1 == EnableMenuItem(hmenu, pos, MF_BYPOSITION | (flag)?MF_ENABLED:MF_DISABLED))
 	  msg_error("Menu item does not exist");
@@ -2656,28 +2751,14 @@ static void msw_get_display_size (int *rows, int *cols)
    *rows = MSW_Screen_Rows = This_Window.height / This_Window.font_height;
 }
 
-static void msw_set_mono (int i, char *what , SLtt_Char_Type unused)
+static JX_SETXXX_RETURN_TYPE msw_set_mono (int i, char *what , SLtt_Char_Type unused)
 {
    (void)i;
    (void)what;
    (void)unused;
+   return JX_SETXXX_RETURN_VAL;
 }
 
-int (*X_Argc_Argv_Hook)(int, char **);
-
-/* the links to functions and variables here */
-void (*tt_beep)(void)  			= msw_beep;
-void (*tt_write_string)(char *)  	= msw_write_string;
-void (*tt_get_terminfo)(void)  		= msw_get_terminfo;
-void (*tt_set_color)(int, char *, char *, char *) = msw_set_color;
-void (*tt_set_mono) (int, char *, SLtt_Char_Type) = msw_set_mono;
-void (*tt_get_screen_size)(int *, int *) = msw_get_display_size;
-
-int *tt_Ignore_Beep  		= &MSW_Ignore_Beep;
-int *tt_Use_Ansi_Colors  	= &MSW_Use_Ansi_Colors;
-int *tt_Term_Cannot_Scroll  	= &MSW_Term_Cannot_Scroll;
-int *tt_Term_Cannot_Insert  	= &MSW_Term_Cannot_Insert;
-
 int jed_init_w32_support (void)
 {
    return w32_init_subprocess_support (1);
@@ -2721,7 +2802,6 @@ static int init_wheel_mouse (void)
 #if defined(__BORLANDC__) || defined(__WIN32__) || defined(__VC__)
 extern int main(int, char **);
 
-/* I do not think that this gets called.  Does it???? */
 int PASCAL WinMain(HINSTANCE inst, HINSTANCE pinst, LPSTR lpszCmdLine, int nCmdShow)
 {
    char **argv;
@@ -2826,3 +2906,80 @@ int PASCAL WinMain(HINSTANCE inst, HINSTANCE pinst, LPSTR lpszCmdLine, int nCmdS
 }
 
 #endif				       /* __BORLANDC__ || __VC__ */
+
+void (*tt_beep)(void);
+void (*tt_write_string)(char *);
+JX_SETXXX_RETURN_TYPE (*tt_set_color)(int, char *, char *, char *);
+JX_SETXXX_RETURN_TYPE (*tt_set_mono) (int, char *, SLtt_Char_Type);
+void (*tt_get_screen_size)(int *, int *);
+int *tt_Ignore_Beep;
+int *tt_Use_Ansi_Colors;
+int *tt_Term_Cannot_Scroll;
+int *tt_Term_Cannot_Insert;
+
+static void init_tt_hooks (void)
+{
+   tt_beep		= msw_beep;
+   tt_write_string	= msw_write_string;
+   tt_get_screen_size	= msw_get_display_size;
+   tt_set_color		= msw_set_color;
+   tt_set_mono		= msw_set_mono;
+
+   tt_Ignore_Beep  	= &MSW_Ignore_Beep;
+   tt_Use_Ansi_Colors  	= &MSW_Use_Ansi_Colors;
+   tt_Term_Cannot_Scroll= &MSW_Term_Cannot_Scroll;
+   tt_Term_Cannot_Insert= &MSW_Term_Cannot_Insert;
+}
+
+static void msw_get_terminfo (void)
+{
+   SLsmg_Term_Type tt;
+
+   init_tt_hooks ();
+     
+   MSW_Screen_Cols = 80;
+   MSW_Screen_Rows = 24;
+
+   (void) jed_add_init_slang_hook (msw_init_slang);
+
+   /* Set this so that main will not try to read from stdin.  It is quite
+    * likely that this is started from a menu or something.
+    */
+   Stdin_Is_TTY = -1;
+   
+   /* init hooks */
+   X_Update_Open_Hook = msw_update_open;
+   X_Update_Close_Hook = msw_update_close;
+   X_Suspend_Hook = msw_suspend;
+   /*   X_Argc_Argv_Hook = X_eval_command_line; */
+   X_Define_Keys_Hook = msw_define_xkeys;
+   SLang_Interrupt = msw_check_kbd;
+   X_Init_Term_Hook = msw_init_term;
+
+   if (Batch)
+     return;
+
+   memset ((char *) &tt, 0, sizeof (SLsmg_Term_Type));
+
+   tt.tt_normal_video = msw_normal_video;
+   tt.tt_set_scroll_region = msw_set_scroll_region;
+   tt.tt_goto_rc = msw_goto_rc;
+   tt.tt_reverse_index = msw_reverse_index;
+   tt.tt_reset_scroll_region = msw_reset_scroll_region;
+   tt.tt_delete_nlines = msw_delete_nlines;
+   tt.tt_cls = msw_cls;
+   tt.tt_del_eol = msw_del_eol;
+   tt.tt_smart_puts = msw_smart_puts;
+   tt.tt_flush_output = msw_flush_output;
+   tt.tt_reset_video = msw_reset_video;
+   tt.tt_init_video = msw_init_video;
+   
+   tt.tt_screen_rows = &MSW_Screen_Rows;
+   tt.tt_screen_cols = &MSW_Screen_Cols;
+   tt.tt_term_cannot_scroll = &MSW_Term_Cannot_Scroll;
+
+   SLsmg_set_terminal_info (&tt);
+}
+
+void (*tt_get_terminfo)(void) = msw_get_terminfo;
+int (*X_Argc_Argv_Hook)(int, char **) = NULL;
